"""
``emsarray`` provides both a command line interface,
and a set of tools to make writing your own command line scripts easier.
"""

import argparse
import importlib
import pkgutil
from typing import Iterable, Type

import emsarray

from . import commands
from .command import BaseCommand
from .exceptions import CommandException
from .utils import console_entrypoint

__all__ = ['main', 'CommandException', 'BaseCommand', 'console_entrypoint']


def command_line_flags(parser: argparse.ArgumentParser) -> None:
    parser.set_defaults(func=lambda x: print(parser.format_help()))
    parser.add_argument(
        '-V', '--version', action='version', version=f'%(prog)s {emsarray.__version__}')

    subparsers = parser.add_subparsers(title="operations", metavar='OPERATION')
    for command_cls in _find_all_commands():
        command_cls().add_parser(subparsers)


@console_entrypoint(command_line_flags)
def main(options: argparse.Namespace) -> None:
    """
    The main entry point for :mod:`emsarray` as a command line utility.

    See also
    --------
    :ref:`cli`
    """
    options.func(options)


def _find_all_commands() -> Iterable[Type[BaseCommand]]:
    for moduleinfo in pkgutil.iter_modules(commands.__path__):
        if moduleinfo.name.startswith('_'):
            continue
        command_path = commands.__name__ + '.' + moduleinfo.name
        command_module = importlib.import_module(command_path)
        yield command_module.Command



"""
Useful functions for writing examples in the documentation
or writing tutorials for new users.
"""

import os
import pathlib
from typing import Any, cast

import xarray

import emsarray
from emsarray.utils import requires_extra

try:
    import pooch
    IMPORT_EXCEPTION = None
except ImportError as exc:
    IMPORT_EXCEPTION = exc


_requires_tutorial = requires_extra('tutorial', IMPORT_EXCEPTION)


BASE_URL = "https://github.com/csiro-coasts/emsarray-data/raw/{version}/{path}"

# Bump this version as new datasets are added or as old datasets are updated.
# Should be a tag name or commit hash, something immutable.
VERSION = "v0.3.0"


def _make_download_url(name: str) -> str:
    """URL to download a named file."""
    return BASE_URL.format(version=VERSION, path=name)


@_requires_tutorial
def _cache_directory() -> pathlib.Path:
    """Where to store cached downloads."""
    if 'EMSARRAY_DATA_DIR' in os.environ:
        path = pathlib.Path(os.environ['EMSARRAY_DATA_DIR'])
    else:
        path = cast(pathlib.Path, pooch.os_cache('emsarray_tutorial') / VERSION)
    path.mkdir(exist_ok=True, parents=True)
    return path


@_requires_tutorial
def _fetch(name: str, cache: bool = True) -> str:
    """
    Fetches a named file from the data repository to the local cache
    and returns the path to it.
    """
    logger = pooch.get_logger()
    logger.setLevel("WARNING")
    if not name.endswith('.nc'):
        name = name + '.nc'
    local_path = cast(str, pooch.retrieve(
        _make_download_url(name),
        known_hash=None,
        path=_cache_directory(),
    ))
    return local_path


@_requires_tutorial
def open_dataset(name: str, **kwargs: Any) -> xarray.Dataset:
    """
    Open one of the example datasets.

    Example datasets will be downloaded from
    the `dataset repository <https://github.com/csiro-coasts/emsarray-data/>`_
    and cached locally.
    Refer to that repository for more information,
    including where the datasets were sourced from
    and full licensing details.

    The example datasets that are available are:

    ``austen``
        A day of the AUSTEn National Tidal model data.
        This dataset is defined on an :class:`unstructured grid <.UGrid>`.

    ``bran2020``
        A small sample of the Bluelink Reanalysis 2020 (BRAN2020) ocean dataset.
        This dataset is defined on a rectangular grid with one dimensional coordinates,
        handled by the :class:`.CFGrid1D` convenion.

    ``fraser``
        A subset of the Great Barrier Reef 4km (GBR4) v2.0 model, part of the eReefs data.
        This subset is centred around K'gari / Fraser Island.
        This dataset is defined on a curvilinear grid with two dimensional coordinates,
        handled by the :class:`.CFGrid2D` convention.
        Temperature, sea surface height, and current variables are included.

    ``gbr4``
        A subset of the Great Barrier Reef 4km (GBR4) v2.0 model, part of the eReefs data.
        This dataset is defined on a curvilinear grid with two dimensional coordinates,
        handled by the :class:`.CFGrid2D` convention.
        Temperature, sea surface heigh, and salinity variables are included.

    Parameters
    ----------
    name : str
        The name of the example dataset to open
    kwargs
        Passed through to :func:`emsarray.open_dataset`.

    Returns
    -------
    :class:`xarray.Dataset`
        The example dataset, after being downloaded and opened by xarray.

    Example
    -------

    .. code-block:: python

        import emsarray
        fraser = emsarray.tutorial.open_dataset('fraser')
        fraser.ems.plot(fraser['temp'].isel(time=0, k=-1))
    """
    local_path = _fetch(name)
    return emsarray.open_dataset(local_path, **kwargs)



"""
Subset a dataset at a set of points.

:func:`.extract_dataframe` takes a pandas :class:`~pandas.DataFrame`,
subsets the dataset at the point specified in each row,
and merges the dataset with the dataframe.
The points extracted will form the coordinates for the new dataset.

:func:`.extract_points` takes a list of Shapely :class:`Points <Point>`,
subsets the dataset at these points,
and returns a new dataset with out any associated geometry.
This is useful if you want to add your own metadata to the subset dataset.

If any of the supplied points does not intersect the dataset geometry,
a :exc:`.NonIntersectingPoints` exception is raised.
This will include the indices of the points that do not intersect.

:ref:`emsarray extract-points` is a command line interface to :func:`.extract_dataframe`.
"""
import dataclasses
from typing import Hashable, List, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry import Point

from emsarray.conventions import Convention


@dataclasses.dataclass
class NonIntersectingPoints(ValueError):
    """
    Raised when a point to extract does not intersect the dataset geometry.
    """

    #: The indices of the points that do not intersect
    indices: np.ndarray

    #: The non-intersecting points
    points: List[Point]

    def __post_init__(self) -> None:
        super().__init__(f"{self.points[0].wkt} does not intersect the dataset geometry")


def _dataframe_to_dataset(
    dataframe: pd.DataFrame,
    *,
    dimension_name: Hashable,
) -> xr.Dataset:
    """
    Convert a pandas DataFrame to an xarray Dataset.
    pandas adds an 'index' coordinate that numbers the 'index' dimension.
    We don't need the coordinate, and the dimension needs to be renamed.
    """
    index_name = dataframe.index.name or 'index'
    dataset = dataframe.to_xarray()
    dataset = dataset.drop_vars(index_name)
    if dimension_name != index_name:
        dataset = dataset.rename_dims({index_name: dimension_name})
    return dataset


def extract_points(
    dataset: xr.Dataset,
    points: List[Point],
    *,
    point_dimension: Hashable = 'point',
) -> xr.Dataset:
    """
    Drop all data except for cells that intersect the given points.
    Return a new dataset with a new dimension named ``point_dimension``,
    with the same size as the nubmer of ``points``,
    containing only data at those points.

    The returned dataset has no coordinate information.

    Parameters
    ----------
    dataset : xarray.Dataset
        The dataset to extract point data from.
    points : list of :class:`Point`
        The points to select.
    point_dimension : Hashable, optional
        The name of the new dimension to index points along.
        Defaults to ``"point"``.

    Returns
    -------
    xarray.Dataset
        A subset of the input dataset that only contains data at the given points.
        The dataset will only contain the values, without any coordinate information.

    See also
    --------
    :func:`extract_dataframe`
    """
    convention: Convention = dataset.ems

    # Find the indexer for each given point
    indexes = np.array([convention.get_index_for_point(point) for point in points])

    # TODO It would be nicer if out-of-bounds points were represented in the
    # output by masked values, rather than raising an error.
    out_of_bounds = np.flatnonzero(np.equal(indexes, None))  # type: ignore
    if len(out_of_bounds):
        raise NonIntersectingPoints(
            indices=out_of_bounds,
            points=[points[i] for i in out_of_bounds])

    # Make a DataFrame out of all point indexers
    selector_df = pd.DataFrame([
        convention.selector_for_index(index.index)
        for index in indexes
        if index is not None])

    # Subset the dataset to the points
    selector_ds = _dataframe_to_dataset(selector_df, dimension_name=point_dimension)
    return convention.drop_geometry().isel(selector_ds)


def extract_dataframe(
    dataset: xr.Dataset,
    dataframe: pd.DataFrame,
    coordinate_columns: Tuple[str, str],
    *,
    point_dimension: Hashable = 'point',
) -> xr.Dataset:
    """
    Extract the points listed in a pandas :class:`~pandas.DataFrame`,
    and merge the remaining columns in to the :class:`~xarray.Dataset`.

    Parameters
    ----------
    dataset : xarray.Dataset
        The dataset to extract point data from
    dataframe : pandas.DataFrame
        A dataframe with longitude and latitude columns,
        and possibly other columns.
    coordinate_columns : tuple of str, str
        The names of the longitude and latitude columns in the dataframe.
    point_dimension : Hashable, optional
        The name of the new dimension to create in the dataset.
        Optional, defaults to "point".

    Returns
    -------
    xarray.Dataset
        A new dataset that only contains data at the given points,
        plus any new columns present in the dataframe.

    Example
    -------

    .. code-block:: python

        import emsarray
        import pandas as pd
        from emsarray.operations import point_extraction

        ds = emsarray.tutorial.open_dataset('gbr4')
        df = pd.DataFrame({
            'lon': [152.807, 152.670, 153.543],
            'lat': [-24.9595, -24.589, -25.488],
            'name': ['a', 'b', 'c'],
        })
        point_data = point_extraction.extract_dataframe(
            ds, df, ['lon', 'lat'])
        point_data

    .. code-block:: text

        <xarray.Dataset>
        Dimensions:  (k: 47, point: 3, time: 1)
        Coordinates:
            zc       (k) float32 ...
          * time     (time) datetime64[ns] 2022-05-11T14:00:00
            lon      (point) float64 152.8 152.7 153.5
            lat      (point) float64 -24.96 -24.59 -25.49
        Dimensions without coordinates: k, point
        Data variables:
            botz     (point) float32 ...
            eta      (time, point) float32 ...
            salt     (time, k, point) float32 ...
            temp     (time, k, point) float32 ...
            name     (point) object 'a' 'b' 'c'
        Attributes: (12/14)
            ...
    """
    lon_coord, lat_coord = coordinate_columns

    # Extract the points from the dataset
    points = [
        Point(row[lon_coord], row[lat_coord])
        for i, row in dataframe.iterrows()]
    point_dataset = extract_points(dataset, points, point_dimension=point_dimension)

    # Merge in the dataframe
    point_dataset = point_dataset.merge(_dataframe_to_dataset(
        dataframe, dimension_name=point_dimension))
    point_dataset = point_dataset.set_coords(coordinate_columns)

    # Add CF attributes to the new coordinate variables
    point_dataset[lon_coord].attrs.update({
        "long_name": "longitude",
        "units": "degrees_east",
        "standard_name": "longitude",
    })
    point_dataset[lat_coord].attrs.update({
        "long_name": "latitude",
        "units": "degrees_north",
        "standard_name": "latitude",
    })

    return point_dataset



"""
Export the geometry of a dataset to a number of formats.
The geometry is represented as a collection of Polygons.
"""
import json
import os
import pathlib
from contextlib import contextmanager
from typing import (
    IO, Any, Generator, Iterable, Iterator, Optional, TypeVar, Union
)

import geojson
import shapefile
import shapely
import xarray as xr

from emsarray.types import Pathish

T = TypeVar('T')


class _dumpable_iterator(list):
    """
    Wrap an iterator / generator so it can be used in `json.dumps()`.
    No guarantees that it works for anything else!

    `json.dumps()` will only iterate over list or tuple subclasses by default,
    and will throw an error if passed an iterator.
    This class pretends to be a list, but presents the iteratable instead.
    Iterating over this multiple times is not supported,
    and most magic methods will access the (empty) list underlying this.
    """
    def __init__(self, gen: Iterable[T]):
        self.gen = gen

    def __iter__(self) -> Iterator[T]:
        return iter(self.gen)

    def __bool__(self) -> bool:
        return True

    def __len__(self) -> int:
        raise NotImplementedError("Can't get the length of a _dumpable_iterator")


def to_geojson(
    dataset: xr.Dataset,
) -> geojson.FeatureCollection:
    """Make a ``geojson.FeatureCollection`` out of the cells in this dataset,
    one feature per cell.

    Each feature will include the linear index and the native index
    of the corresponding cell in its properties.

    .. note::
        The `features` list on the returned FeatureCollection is a generator.
        It can only be iterated over once.
        It is designed to work with :func:`json.dumps` directly.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset to export as GeoJSON geometry.

    Returns
    -------
    ``geojson.FeatureCollection``
        The geometry of this dataset as a FeatureCollection.

    See also
    --------
    :func:`.write_geojson`
    """
    return geojson.FeatureCollection(_dumpable_iterator(
        geojson.Feature(geometry=polygon, properties={
            'linear_index': i,
            'index': dataset.ems.unravel_index(i),
        })
        for i, polygon in enumerate(dataset.ems.polygons)
        if polygon is not None
    ))


def write_geojson(
    dataset: xr.Dataset,
    path: Pathish,
) -> None:
    """
    Export the geometry of this dataset to a GeoJSON file
    as a ``geojson.FeatureCollection`` with one feature per cell.

    Each feature will include the linear index and the native index
    of the corresponding cell in its properties.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset to export as GeoJSON geometry.
    path : str or pathlib.Path
        The path where the geometry should be written to.

    See also
    --------
    :func:`.to_geojson`
    """
    with open(path, 'w') as f:
        json.dump(to_geojson(dataset), f)


@contextmanager
def _maybe_open(path_or_file: Union[Pathish, IO], mode: str) -> Generator[IO, None, None]:
    """
    Given either a path to a file or an open file handle,
    return an open file handle wrapped in a context manager.

    If given a path, the file is opened and returned as the context manager.
    The file is closed when the context manager closes.

    If given an open file handle, the handle is returned as-is.
    The file is not closed when the context manager closes.
    """
    if isinstance(path_or_file, IO):
        yield path_or_file
    else:
        with open(path_or_file, mode) as f:
            yield f


def write_shapefile(
    dataset: xr.Dataset,
    target: Optional[Pathish] = None,
    *,
    shp: Optional[Union[Pathish, IO]] = None,
    shx: Optional[Union[Pathish, IO]] = None,
    dbf: Optional[Union[Pathish, IO]] = None,
    prj: Optional[Union[Pathish, IO]] = None,
    **kwargs: Any,
) -> None:
    """
    Write the geometry of this dataset to a Shapefile.
    Each polygon is saved as an individual record
    with ``name``, ``linear_index``, and ``index`` fields.

    Parameters
    ----------
    dataset : xarray.Dataset
        The dataset to export the geometry for.
    target : str or pathlib.Path, optional
        Where to save the shapefile.
        The names for the .shp, .shx, .dbf, and .prj files are derived from this.
        Optional, you can instead provide arguments for each individual file.
    shp, shx, dbf, prj : str or pathlib.Path or opened file handle
        Where to write the individual shapefile components to.
        Optional, you can instead provide the base path as ``target``.
    **kwargs
        An extra keyword arguments are passed on to the shapefile.Writer instance.
    """
    target = str(target) if isinstance(target, pathlib.Path) else target
    with shapefile.Writer(target, shp=shp, shx=shx, dbf=dbf, **kwargs) as writer:
        writer.field('name', 'C')
        writer.field('linear_index', 'N')
        writer.field('index', 'C')
        for i, polygon in enumerate(dataset.ems.polygons):
            if polygon is None:
                continue
            writer.record(
                name=f'polygon{i}',
                linear_index=i,
                index=json.dumps(dataset.ems.unravel_index(i)),
            )
            writer.shape(polygon.__geo_interface__)

        # Write the projection file also, if we can find a filename for it...
        if prj is None:
            if target is not None:
                prj = os.path.splitext(target)[0] + '.prj'
        if prj is not None:
            with _maybe_open(prj, 'w') as prj_file:
                prj_file.write(dataset.ems.data_crs.to_wkt())


def _to_multipolygon(dataset: xr.Dataset) -> shapely.MultiPolygon:
    return shapely.MultiPolygon([
        p for p in dataset.ems.polygons
        if p is not None
    ])


def write_wkt(
    dataset: xr.Dataset,
    path: Pathish,
) -> None:
    """
    Export the geometry of this dataset as a ``MultiPolygon``
    to a Well Known Text file.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset to export as Well Known Text.
    path : str or pathlib.Path
        The path where the geometry should be written to.
    """
    with open(path, 'w') as f:
        f.write(shapely.to_wkt(_to_multipolygon(dataset)))


def write_wkb(
    dataset: xr.Dataset,
    path: Pathish,
) -> None:
    """
    Export the geometry of this dataset as a ``MultiPolygon``
    to a Well Known Binary file.

    Parameters
    ----------
    dataset : xr.Dataset
        The dataset to export as Well Known Binary.
    path : str or pathlib.Path
        The path where the geometry should be written to.
    """
    with open(path, 'wb') as f:
        f.write(shapely.to_wkb(_to_multipolygon(dataset)))



"""
emsarray - xarray utilities for various non-CF grid data conventions
"""
import importlib.metadata
import warnings
from typing import Any

from . import tutorial
from .accessors import ems_accessor
from .conventions import Convention, get_dataset_convention, open_dataset

__version__ = importlib.metadata.version("emsarray")

__all__ = [
    "tutorial",
    "ems_accessor",
    "Convention", "get_dataset_convention", "open_dataset",
]


def __getattr__(name: str) -> Any:
    old_new = {
        'Format': ('Convention', 'emsarray.formats.Format'),
        'get_file_format': ('get_dataset_convention', 'emsarray.formats.get_file_format'),
    }
    if name in old_new:
        new, old_path = old_new[name]
        warnings.warn(
            f"emsarray.{name} has been renamed to emsarray.{new}",
            category=DeprecationWarning,
            stacklevel=2)
        old_module, old_name = old_path.rsplit('.', 1)
        return getattr(importlib.import_module(old_module), old_name)

    raise AttributeError(f"{name} is not defined")



import subprocess
from pathlib import Path
from unittest import mock

import pytest

from emsarray import nco


def test_ncrcat(monkeypatch: pytest.MonkeyPatch) -> None:
    check_call = mock.MagicMock(subprocess.check_call)
    monkeypatch.setattr(subprocess, 'check_call', check_call)

    nco.ncrcat(["one.nc", "two.nc"], "out.nc")
    check_call.assert_called_once_with(
        ['ncrcat', '-hH', '-o', 'out.nc', 'one.nc', 'two.nc'],
        stdin=subprocess.DEVNULL,
    )


def test_ncrcat_output_exists(tmpdir: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    out_file = Path(tmpdir) / 'out.nc'
    out_file.touch()

    check_call = mock.MagicMock(subprocess.check_call)
    monkeypatch.setattr(subprocess, 'check_call', check_call)

    with pytest.raises(FileExistsError):
        nco.ncrcat(["one.nc", "two.nc"], out_file)

    check_call.assert_not_called()


def test_ncrcat_no_input_files(monkeypatch: pytest.MonkeyPatch) -> None:
    check_call = mock.MagicMock(subprocess.check_call)
    monkeypatch.setattr(subprocess, 'check_call', check_call)

    with pytest.raises(ValueError):
        nco.ncrcat([], "out.nc")

    check_call.assert_not_called()



import argparse
import contextlib
import logging
import tempfile
from pathlib import Path
from typing import ContextManager

import emsarray
from emsarray.cli import BaseCommand
from emsarray.cli.utils import geometry_argument
from emsarray.types import Pathish

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Clip a dataset to the given geographic bounds"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "input_path", type=Path,
            help="Path to input netCDF4 file")
        parser.add_argument(
            'clip_geometry', type=geometry_argument, metavar='clip',
            help=(
                "Shape to clip the dataset to. "
                "Can be a bounding box defined as 'lon_min,lat_min,lon_max,lat_max'; "
                "a WKT string; a geojson string; or a path to a geojson file."
            ))
        parser.add_argument(
            "output_path", type=Path,
            help="Path to output netCDF4 file")

        parser.add_argument(
            "--work_dir", type=Path, default=None,
            help=(
                "Where to put temporary files. If not provided, "
                "the system temporary directory will be used."
            ))

    def handle(self, options: argparse.Namespace) -> None:
        work_context: ContextManager[Pathish]
        if options.work_dir:
            work_context = contextlib.nullcontext(options.work_dir)
        else:
            work_context = tempfile.TemporaryDirectory(prefix="emsarray-clip.")

        logger.info("Clipping %r", str(options.input_path))
        dataset = emsarray.open_dataset(options.input_path)
        with work_context as work_path:
            logger.debug("Using %r as work dir", str(work_path))
            out = dataset.ems.clip(options.clip_geometry, work_dir=work_path)
            logger.debug("Saving to %r", str(options.output_path))
            out.ems.to_netcdf(options.output_path)
        logger.info("Saved to %r", str(options.output_path))



class CommandException(RuntimeError):
    """Raise this to exit a command line script gracefully.

    See also
    --------
    :func:`.utils.console_entrypoint`
    """
    def __init__(self, message: str, code: int = 1):
        self.message = message
        self.code = code



import pytest
import xarray as xr

import emsarray
from emsarray.conventions import get_dataset_convention
from emsarray.conventions.grid import CFGrid1D, CFGrid2D
from emsarray.conventions.shoc import ShocSimple, ShocStandard
from emsarray.conventions.ugrid import UGrid


@pytest.mark.parametrize(['dataset', 'expected_class'], [
    ('cfgrid1d.nc', CFGrid1D),
    ('cfgrid2d.nc', CFGrid2D),
    ('shoc_standard.nc', ShocStandard),
    ('ugrid_mesh2d.nc', UGrid),
])
def test_files(datasets, dataset, expected_class):
    dataset = xr.open_dataset(datasets / dataset)
    assert get_dataset_convention(dataset) is expected_class


@pytest.mark.tutorial
@pytest.mark.parametrize(['dataset', 'expected_class'], [
    ('bran2020', CFGrid1D),
    ('gbr4', ShocSimple),
    ('austen', UGrid),
])
def test_tutorial_datasets(dataset, expected_class):
    dataset = emsarray.tutorial.open_dataset(dataset)
    assert get_dataset_convention(dataset) is expected_class



import abc
import argparse
from typing import Optional

from emsarray.cli import utils


class BaseCommand(abc.ABC):
    """
    Base class for writing :mod:`emsarray` command line tools.
    Subclasses of this can be added to the :mod:`emsarray` entry point :func:`.main`.
    """

    @property
    def name(self) -> str:
        """
        The name of this subcommand.
        Users can invoke this subcommand by running ``emsarray <name>``.
        Defaults to the name of the module,
        with underscores replaced with hyphens.
        """
        module_name = type(self).__module__.rsplit('.', 1)[-1]
        return module_name.replace('_', '-')

    #: A short description of what this subcommand does,
    #: shown as part of the usage message for the base command.
    help: Optional[str] = None

    #: A longer description of what this subcommand does,
    #: shown as part of the usage message for this subcommand.
    description: Optional[str] = None

    def add_parser(self, subparsers: argparse._SubParsersAction) -> None:
        parser = subparsers.add_parser(
            name=self.name,
            help=self.help,
            description=self.description,
            formatter_class=utils.DoubleNewlineDescriptionFormatter,
        )
        parser.set_defaults(func=self.handle)
        self.add_arguments(parser)

    @abc.abstractmethod
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Configure the :class:`argparse.ArgumentParser` for this command,
        adding any flags and options required.
        """
        ...

    @abc.abstractmethod
    def handle(self, options: argparse.Namespace) -> None:
        """
        Run this command. ``options`` will be the parsed command line flags,
        as configured in :meth:`.add_arguments`.
        """
        ...



"""
Python aliases to ``nco`` utilities.

:mod:`xarray` is great, but can not handle large files in the same way that
``nco`` tools can. This module provides a set of Python functions that call
``nco`` tools, for times when :mod:`xarray` is not sufficient.

All these functions take paths to datasets as arguments, not
:class:`xarray.Dataset` instances.
"""
import subprocess
from pathlib import Path
from typing import Any, Optional, Sequence, Union

Pathish = Union[Path, str]


def _check_call(cmd: Sequence[str], stdin: Any = subprocess.DEVNULL, **kwargs: Any) -> None:
    subprocess.check_call(cmd, stdin=stdin, **kwargs)


def ncrcat(
    input_files: Sequence[Pathish],
    output_file: Pathish,
    flags: Optional[str] = None,
    history: bool = False,
) -> None:
    """Concatenates a set of netCDF files together using `ncrcat`."""
    if Path(output_file).exists():
        raise FileExistsError(f"Output file {str(output_file)!r} already exists!")
    if len(input_files) == 0:
        raise ValueError("Input file list is empty")

    cmd = ['ncrcat']
    if not history:
        cmd += ['-hH']
    if flags:
        cmd += flags
    cmd += ['-o', str(output_file)]
    cmd.extend(map(str, input_files))
    _check_call(cmd)



"""
Common functions for working with dataset masks.
Masks are used when clipping datasets to a smaller geographic subset,
such as :meth:`.Convention.clip`.
"""
from __future__ import annotations

import functools
import itertools
import logging
import operator
import pathlib
from typing import Any, Dict, Hashable, List, cast

import numpy as np
import xarray as xr

from emsarray import utils
from emsarray.types import Pathish

logger = logging.getLogger(__name__)


def mask_grid_dataset(
    dataset: xr.Dataset,
    mask: xr.Dataset,
    work_dir: Pathish,
    **kwargs: Any,
) -> xr.Dataset:
    """Apply a mask to a two-dimensional grid dataset,
    such as :class:`.CFGrid1D` and :class:`.CFGrid2D`,
    or datasets with multiple grids such as :class:`.ArakawaC`

    Parameters
    ----------
    dataset
        The :class:`~xarray.Dataset` instance to mask
    mask
        The mask to apply. Different types of datasets need different masks.
    work_dir
        An empty directory where temporary files can be stored while applying
        the mask. The returned dataset will be built from files inside this
        directory, so callers must save the returned dataset before deleting
        this directory.
    kwargs
        Any extra kwargs are passed to `open_mfdataset` when assembling the
        new, clipped dataset.

    Returns
    -------
    :class:`~xarray.Dataset`
        The masked dataset
    """
    work_path = pathlib.Path(work_dir)

    logger.info("Calculating mask bounds...")
    # Trim the dataset and mask down to the new bounds. Doing this before
    # applying the mask is considerably faster
    bounds = calculate_grid_mask_bounds(mask)
    logger.debug("Bounds: %r", bounds)

    logger.info("Slicing mask and dataset to bounds...")
    mask = mask.isel(bounds)
    dataset = dataset.isel(bounds)

    mfdataset_names: List[pathlib.Path] = []

    logger.info("Applying masks...")
    # This is done variable-by-variable, as trying to do it to the entire
    # dataset is very memory intensive. This allows us to offload data to the
    # file system, at the added expense of having to recombine the dataset
    # afterwards.
    for key, data_array in dataset.data_vars.items():
        logger.debug("DataArray %s", key)
        masked_data_array = mask_grid_data_array(mask, data_array)
        variable_path = work_path / f"{key}.nc"
        mfdataset_names.append(variable_path)
        utils.to_netcdf_with_fixes(masked_data_array.to_dataset(name=key), variable_path)
        del masked_data_array

    # Coordinates are handled differently. They have been trimmed already, but
    # shouldn't be masked. They are combined in to one dataset and saved as-is
    coords_path = work_path / "__coords__.nc"
    mfdataset_names.append(coords_path)
    utils.to_netcdf_with_fixes(xr.Dataset(coords=dataset.coords), coords_path)

    logger.info("Merging individual variables")
    # Happily `mfdataset` will load data in to memory in a lazy manner,
    # allowing us to combine very large datasets without running out of memory.
    merged_dataset = xr.open_mfdataset(
        mfdataset_names,
        # `lock=False` prevents an issue where opening a dataset with
        # `open_mfdataset` then saving it with `.to_netcdf()` would
        # intermittently hang.
        #
        # See https://github.com/pydata/xarray/issues/3961
        lock=False,
        **kwargs,
    )

    return utils.dataset_like(dataset, merged_dataset)


def mask_grid_data_array(mask: xr.Dataset, data_array: xr.DataArray) -> xr.DataArray:
    """
    Apply a mask to a single data array.
    A mask dataset contains one or more mask data arrays.
    The mask to apply is selected by comparing dimensions
    - the first mask found which has dimensions
    that are a subset of the data array dimensions is used.

    Parameters
    ----------
    mask : xarray.Dataset
        The mask dataset
    data_array : xarray.DataArray
        The :class:`~xarray.DataArray` to mask

    Returns
    -------
    xarray.DataArray
        A new :class:`~xarray.DataArray`
        with any masked values replaced with _FillValue.
        The returned data array will be the same shape as the input data array.
        If no appropriate mask is found,
        the original data array is returned unmodified.
    """
    dimensions = set(data_array.dims)

    try:
        fill_value = find_fill_value(data_array)
    except ValueError:
        return data_array

    # Loop through each possible mask
    for mask_name, mask_data_array in mask.data_vars.items():
        # If every dimension of this mask exists in the data array, apply it
        if dimensions >= set(mask_data_array.dims):
            new_data_array = cast(xr.DataArray, data_array.where(mask_data_array, other=fill_value))
            new_data_array.attrs = data_array.attrs
            new_data_array.encoding = data_array.encoding
            return new_data_array

    # Fallback, no appropriate mask was found, so don't apply any.
    # This generally happens for data arrays such as time, record, x_grid, etc.
    return data_array


def find_fill_value(data_array: xr.DataArray) -> Any:
    """
    Float-typed variables can easily be masked. If they don't already have
    a fill value, they can be masked using `NaN` without issue.
    However there are some `int`-typed variables without a fill value that
    _cant_ be automatically masked.

    Parameters
    ----------
    data_array : xarray.DataArray
        The :class:`~xarray.DataArray` to find an appropriate ``_FillValue`` for.

    Returns
    -------
    fill value
        A numpy scalar value appropriate to use
        as the fill value in the data array.

        * For masked arrays, this will be :data:`numpy.ma.masked` ---
          note that xarray itself does not use masked arrays, but is compatible with them.
        * For data arrays that already have a ``_FillValue`` used by xarray,
          :data:`numpy.nan` is returned.
          xarray will substitute in all ``_FillValue``
          with :data:`numpy.nan` when opening files.
        * For data arrays that have been opened with ``mask_and_scale=False``,
          the existing ``_FillValue`` is returned.
        * If the data array has a float ``dtype``,
          :data:`numpy.nan` is returned.
        * If none of the above are true, a :exc:`ValueError` is raised.

    """
    if np.ma.is_masked(data_array.values):
        # xarray does not use masked arrays, but just in case someone has
        # constructed a dataset using one...
        return np.ma.masked

    if '_FillValue' in data_array.encoding:
        # The dataset was opened with mask_and_scale=True and a mask has been
        # applied. Masked values are now represented as np.nan, not _FillValue.
        return np.nan

    if '_FillValue' in data_array.attrs:
        # The dataset was opened with mask_and_scale=False and a mask has not
        # been applied. Masked values should be represented using _FillValue.
        return data_array.attrs['_FillValue']

    if issubclass(data_array.dtype.type, np.floating):
        # NaN is a useful fallback for a _FillValue, but only if the dtype
        # is some sort of float. We won't actually _set_ a _FillValue
        # attribute though, as that can play havok when trying to save
        # existing datasets. xarray gets real grumpy when you have
        # a _FillValue and a missing_value, and some existing datasets play
        # fast and loose with mixing the two.
        return np.nan

    raise ValueError("No appropriate fill value found")


def calculate_grid_mask_bounds(mask: xr.Dataset) -> Dict[Hashable, slice]:
    """
    Calculate the included bounds of a mask dataset for each dimension.

    Parameters
    ----------
    mask : xarray.Dataset
        The mask dataset should contain one or more boolean data arrays.

    Returns
    -------
    dict
        A dict of ``{dimension_name: slice(min_index, max_index)}`` will be returned.
        This dict can be passed directly in to a :meth:`xarray.Dataset.isel` call
        to crop a dataset to the bounds of a mask.
    """
    bounds = {}
    for name, mask_data_array in mask.data_vars.items():
        # Assert there is at least one true value somewhere in the entire mask
        # If this is False, strange errors would happen.
        if not mask_data_array.any().item():
            raise ValueError(f"Mask {name!r} is completely empty!")

        # Find the bounds for each dimension
        dimensions_set = set(mask_data_array.dims)
        for dimension in mask_data_array.dims:
            # For each step along this dimension, see if there are any True
            # cells in any of the other dimensions
            values = mask_data_array.any(dim=list(dimensions_set - {dimension}))

            # Find the first and last True values
            min_index = next(i for i, value in enumerate(values) if value)
            max_index = next(len(values) - i for i, value in enumerate(reversed(values)) if value)

            # max_index will actually be the index one _past_ the final True
            # value, but as slice ends are not inclusive, this is the desired
            # behaviour.
            bounds[dimension] = slice(min_index, max_index)

    return bounds


def smear_mask(arr: np.ndarray, pad_axes: List[bool]) -> np.ndarray:
    """
    Take a boolean numpy array and a list indicating which axes to smear along.
    Return a new array, expanded along the axes, with the boolean values
    smeared accordingly.

    This is a half baked convolution operator where the `pad_axes` parameter is
    used to build the kernel.

    Parameters
    ----------
    arr : numpy.ndarray
        A boolean numpy :class:`numpy.ndarray`.
    pad_axes : list of bool
        A list of booleans, indicating which axes to smear along.

    Returns
    -------
    numpy.ndarray
        The smeared array.
        For every axis where ``pad_axes`` was True,
        the array will be one element larger.

    Examples
    --------

    .. code-block:: pycon

        >>> arr
        array([[0, 0, 1, 0, 0],
               [0, 1, 0, 1, 0],
               [1, 0, 0, 0, 1]]

    Smear along the y-axis:

    .. code-block:: pycon

        >>> smear_mask(arr, [False, True])
        array([[0, 0, 1, 1, 0, 0],
               [0, 1, 1, 1, 1, 0],
               [1, 1, 0, 0, 1, 1]]

    Smear along both axes:

    .. code-block:: pycon

        >>> smear_mask(arr, [True, True])
        array([[0, 0, 1, 1, 0, 0],
               [0, 1, 1, 1, 1, 0],
               [1, 1, 1, 1, 1, 1],
               [1, 1, 0, 0, 1, 1])
    """
    paddings = itertools.product(*(
        [(1, 0), (0, 1)] if pad_axis else [(0, 0)]
        for pad_axis in pad_axes
    ))
    return functools.reduce(operator.or_, (np.pad(arr, pad) for pad in paddings))


def blur_mask(arr: np.ndarray, size: int = 1) -> np.ndarray:
    """
    Take a boolean numpy array and blur it, such that all indices neighbouring
    a True value in the input array are True in the output array. The output
    array will have the same shape as the input array.

    Parameters
    ----------
    arr : numpy.ndarray
        A boolean array to blur
    size : int
        The kernel size to use when blurring.
        In the output array,
        any cell that has a true value
        within its `size` neighbours along any axis
        is true.

    Returns
    -------
    numpy.ndarray
        The blurred array

    Examples
    --------

    .. code-block:: python

        >>> arr = numpy.array([
        ...     [1, 0, 0, 0, 0],
        ...     [0, 0, 0, 0, 0],
        ...     [0, 0, 0, 1, 0],
        ...     [0, 0, 0, 0, 1]])
        >>> blur_mask(arr)
        array([[1, 1, 0, 0, 0],
               [1, 1, 1, 1, 1],
               [0, 0, 1, 1, 1],
               [0, 0, 1, 1, 1]]
    """
    # Pad the mask with a `size` sized buffer.
    # This allows simple slicing to pull out a rectangular region around an
    # index, without worrying about the edges of the array.
    padded = np.pad(arr, size, constant_values=False)

    # For each cell in the original mask shape,
    # the blurred mask is true if the original mask was true,
    # or any cells in a `size` sized slice around the original cell.
    arr_iter = np.nditer(arr, ['multi_index'])
    indices = (arr_iter.multi_index for _ in arr_iter)
    values = (
        arr[index] or np.any(padded[tuple(slice(i, i + size * 2 + 1) for i in index)])
        for index in indices
    )

    arr = np.fromiter(values, count=arr.size, dtype=arr.dtype).reshape(arr.shape)
    return cast(np.ndarray, arr)



"""
Curvilinear Arakawa C grids.

See also
--------
`Arakawa grids <https://en.wikipedia.org/wiki/Arakawa_grids>`_ on Wikipedia

"""
from __future__ import annotations

import enum
import logging
from functools import cached_property
from typing import Dict, Hashable, Optional, Tuple, cast

import numpy as np
import xarray as xr
from shapely.geometry.base import BaseGeometry
from xarray.core.dataset import DatasetCoordinates

from emsarray import masking, utils
from emsarray.types import Pathish

from ._base import Convention, Specificity

logger = logging.getLogger(__name__)


class ArakawaCGridTopology:
    """
    A topology helper to deal with grid coordinates.

    Parameters
    ----------
    dataset : xarray.Dataset
        The Arakawa C gridded :class:`~xarray.Dataset` to introspect.
    latitude : Hashable
        The name of the latitude coordinate variable.
    longitude : Hashable
        The name of the longitude coordinate variable.
    """
    def __init__(
        self, dataset: xr.Dataset, *, latitude: Hashable, longitude: Hashable
    ) -> None:
        self.dataset = dataset
        self.latitude_name = latitude
        self.longitude_name = longitude

    @cached_property
    def latitude(self) -> xr.DataArray:
        """The latitude :class:`~xarray.DataArray` coordinate variable."""
        return self.dataset[self.latitude_name]

    @cached_property
    def longitude(self) -> xr.DataArray:
        """The logitude :class:`~xarray.DataArray` coordinate variable."""
        return self.dataset[self.longitude_name]

    @cached_property
    def j_dimension(self) -> Hashable:
        """The name of the ``j`` dimension for this grid kind."""
        return self.latitude.dims[0]

    @cached_property
    def i_dimension(self) -> Hashable:
        """The name of the ``i`` dimension for this grid kind."""
        return self.latitude.dims[1]

    @cached_property
    def shape(self) -> Tuple[int, int]:
        """The shape of this grid, as a tuple of ``(j, i)``."""
        return (self.dataset.dims[self.j_dimension], self.dataset.dims[self.i_dimension])

    @cached_property
    def size(self) -> int:
        """The size of this grid, ``j * i``."""
        return cast(int, np.prod(self.shape))

    def __repr__(self) -> str:
        bits = (f"{key}: {value!r}" for key, value in [
            ('latitude', self.latitude_name),
            ('longitude', self.longitude_name),
            ('shape', self.shape),
        ])
        details = ', '.join(bits)
        name = type(self).__name__
        return f'<{name}: {details}>'


class ArakawaCGridKind(str, enum.Enum):
    """Araawa C grid datasets can store data on
    cell faces, left edges, back edges, and nodes.
    The kind of grid is specified by this enum.
    """
    #: The face grid.
    #:
    #: :meta hide-value:
    face = 'face'
    #: The left edge grid.
    #: The left edge *i* dimension is one larger than for the face *i* dimension.
    #: A face ``(j, i)`` will have left edges ``(j, i)`` and ``(j, i + 1)``
    #:
    #: :meta hide-value:
    left = 'left'
    #: The back edge grid.
    #: The back edge *j* dimension is one larger than for the face *j* dimension.
    #: A face ``(j, i)`` will have back edges ``(j, i)`` and ``(j + 1, i)``
    #:
    #: :meta hide-value:
    back = 'back'
    #: The node grid.
    #: The node *i* and *j* dimensions are one larger than for the face *i* and *j* dimensions.
    #: A face ``(j, i)`` will have nodes ``(j, i)``, ``(j + 1, i)``, ``(j, i + 1``, and ``(j + 1, i + 1)``
    #:
    #: :meta hide-value:
    node = 'node'

    def __call__(self, j: int, i: int) -> ArakawaCIndex:
        return (self, j, i)


#: The native index type for Arakawa C grids
#: is a tuple with three elements: ``(kind, j, i).``
#:
#: :meta hide-value:
ArakawaCIndex = Tuple[ArakawaCGridKind, int, int]
ArakawaCCoordinates = Dict[ArakawaCGridKind, Tuple[Hashable, Hashable]]
ArakawaCDimensions = Dict[ArakawaCGridKind, Tuple[Hashable, Hashable]]


class ArakawaC(Convention[ArakawaCGridKind, ArakawaCIndex]):
    """
    An Arakawa C grid is a curvilinear orthogonal grid
    with data defined on grid faces, edges, and nodes.
    The edges are split in to left and back edges.

    There is a topology helper for each of the
    :attr:`ArakawaC.face`, :attr:`ArakawaC.left`, :attr:`.back`, and :attr:`.node` grids.

    Parameters
    ----------
    dataset : xarray.Dataset
        The dataset containing the Arakawa C grid
    coordinate_names : dict
        A dict defining which coordinate variables define which grids.
        The keys are each of 'face', 'left', 'back', or 'node';
        and the values are a two-tuple of strings
        naming the *latitude* and *longitude* coordinate variables for that grid kind:

        .. code-block:: python

            coordinate_names = {
                'face': ('y_face', 'x_face'),
                'left': ('y_left', 'x_left'),
                'back': ('y_back', 'x_back'),
                'node': ('y_node', 'x_node'),
            }

    """
    grid_kinds = frozenset(ArakawaCGridKind)
    default_grid_kind = ArakawaCGridKind.face

    coordinate_names: ArakawaCCoordinates

    def __init__(
        self,
        dataset: xr.Dataset,
        *,
        coordinate_names: Optional[Dict[Hashable, Tuple[Hashable, Hashable]]] = None,
    ):
        super().__init__(dataset)

        if coordinate_names is not None:
            if set(coordinate_names.keys()) != set(ArakawaCGridKind):
                kinds = ", ".join(kind.name for kind in ArakawaCGridKind)
                raise ValueError(f"coordinate_names must have keys {kinds}")

            self.coordinate_names = {
                ArakawaCGridKind(kind): coords
                for kind, coords in coordinate_names.items()
            }

        # Check that coordinate_names has been defined.
        # It may have been hard coded on a subclass.
        if not hasattr(self, 'coordinate_names'):
            raise TypeError(
                "Grid coordinates have not been defined. "
                "You must either pass the `coordinate_names` parameter, "
                "or set it on a subclass."
            )

    @cached_property
    def _dimensions_for_grid_kind(self) -> ArakawaCDimensions:
        return {
            kind: cast(Tuple[Hashable, Hashable], self.dataset[coordinates[0]].dims)
            for kind, coordinates in self.coordinate_names.items()
        }

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        if not hasattr(cls, 'coordinate_names'):
            return None

        if all(
            coord in dataset.variables
            for coords in cls.coordinate_names.values() for coord in coords
        ):
            return Specificity.HIGH

        return None

    @cached_property
    def _topology_for_grid_kind(self) -> Dict[ArakawaCGridKind, ArakawaCGridTopology]:
        return {
            kind: ArakawaCGridTopology(
                self.dataset,
                latitude=coords[0],
                longitude=coords[1],
            )
            for kind, coords in self.coordinate_names.items()
        }

    @cached_property
    def face(self) -> ArakawaCGridTopology:
        """
        Topology for the face grid,
        an instance of :class:`ArakawaCGridTopology`.
        """
        return self._topology_for_grid_kind[ArakawaCGridKind.face]

    @cached_property
    def left(self) -> ArakawaCGridTopology:
        """
        Topology for the left edge grid,
        an instance of :class:`ArakawaCGridTopology`.
        """
        return self._topology_for_grid_kind[ArakawaCGridKind.left]

    @cached_property
    def back(self) -> ArakawaCGridTopology:
        """
        Topology for the back edge grid,
        an instance of :class:`ArakawaCGridTopology`.
        """
        return self._topology_for_grid_kind[ArakawaCGridKind.back]

    @cached_property
    def node(self) -> ArakawaCGridTopology:
        """
        Topology for the node grid,
        an instance of :class:`ArakawaCGridTopology`.
        """
        return self._topology_for_grid_kind[ArakawaCGridKind.node]

    def unravel_index(
        self,
        index: int,
        grid_kind: Optional[ArakawaCGridKind] = None,
    ) -> ArakawaCIndex:
        if grid_kind is None:
            grid_kind = ArakawaCGridKind.face
        topology = self._topology_for_grid_kind[grid_kind]
        j, i = map(int, np.unravel_index(index, topology.shape))
        return (grid_kind, j, i)

    def ravel_index(self, indices: ArakawaCIndex) -> int:
        grid_kind, j, i = indices
        topology = self._topology_for_grid_kind[grid_kind]
        return int(np.ravel_multi_index((j, i), topology.shape))

    def get_grid_kind_and_size(self, data_array: xr.DataArray) -> Tuple[ArakawaCGridKind, int]:
        dims = set(data_array.dims)
        for grid_kind in ArakawaCGridKind:
            grid_kind_dims = self._dimensions_for_grid_kind[grid_kind]
            if dims.issuperset(grid_kind_dims):
                topology = self._topology_for_grid_kind[grid_kind]
                return grid_kind, topology.size

        raise ValueError("Data array did not match any known grids")

    @cached_property
    def polygons(self) -> np.ndarray:
        # Make an array of shape (j, i, 2) of all the nodes
        grid = np.stack([self.node.longitude.values, self.node.latitude.values], axis=-1)

        # Transform this in to an array of shape (topology.size, 4, 2)
        points = np.stack([
            grid[:-1, :-1],
            grid[:-1, +1:],
            grid[+1:, +1:],
            grid[+1:, :-1],
        ], axis=2).reshape((-1, 4, 2))

        return utils.make_polygons_with_holes(points)

    @cached_property
    def face_centres(self) -> np.ndarray:
        centres = np.column_stack((
            self.make_linear(self.face.longitude).values,
            self.make_linear(self.face.latitude).values,
        ))
        return cast(np.ndarray, centres)

    def selector_for_index(self, index: ArakawaCIndex) -> Dict[Hashable, int]:
        kind, j, i = index
        topology = self._topology_for_grid_kind[kind]
        return {topology.j_dimension: j, topology.i_dimension: i}

    def drop_geometry(self) -> xr.Dataset:
        variables = [
            self.face.longitude.name,
            self.face.latitude.name,
            self.node.longitude.name,
            self.node.latitude.name,
            self.left.longitude.name,
            self.left.latitude.name,
            self.back.longitude.name,
            self.back.latitude.name,
        ]
        return self.dataset.drop_vars(variables)

    def make_linear(self, data_array: xr.DataArray) -> xr.DataArray:
        kind, size = self.get_grid_kind_and_size(data_array)
        topology = self._topology_for_grid_kind[kind]
        dimensions = [topology.j_dimension, topology.i_dimension]
        return utils.linearise_dimensions(data_array, list(dimensions))

    def make_clip_mask(
        self,
        clip_geometry: BaseGeometry,
        buffer: int = 0,
    ) -> xr.Dataset:
        """
        Generate an xarray.Dataset with a mask for the cell centres, left and
        back faces, and nodes. The mask values will be True where the
        coordinate intersects the clip polygon, and False elsewhere.
        """
        logger.info("Finding intersecting cells for centre mask")

        # A cell is included if it intersects the clip polygon
        intersecting_indices = [
            item.linear_index
            for polygon, item in self.spatial_index.query(clip_geometry)
            if polygon.intersects(clip_geometry)]
        face_mask = np.full(self.face.shape, fill_value=False)
        face_mask.ravel()[intersecting_indices] = True

        # Expand the mask by one cell around the clipped region, as a buffer
        if buffer > 0:
            face_mask = masking.blur_mask(face_mask, size=buffer)

        # Complete the rest of the mask
        return c_mask_from_centres(face_mask, self._dimensions_for_grid_kind, self.dataset.coords)

    def apply_clip_mask(self, clip_mask: xr.Dataset, work_dir: Pathish) -> xr.Dataset:
        return masking.mask_grid_dataset(self.dataset, clip_mask, work_dir)


def c_mask_from_centres(
    face_mask: np.ndarray,
    dimensions: ArakawaCDimensions,
    coords: Optional[DatasetCoordinates] = None,
) -> xr.Dataset:
    """
    Create a mask for a SHOC standard file given a mask array for the cell
    centres to include. The full mask will include the centres, edges, and nodes.
    """
    # If a cell intersects the clip region at all, it will be included. This means
    # we need to include the cell centres, and also the edges for that cell.
    # After finding the centre mask, we can carefully pad that array to find the
    # left edge mask, back edge mask, and node masks.

    # An edge is included if the cell on either side is included.
    left_mask = masking.smear_mask(face_mask, [False, True])
    back_mask = masking.smear_mask(face_mask, [True, False])

    # A node is included if any of the four surrounding cells are included.
    node_mask = masking.smear_mask(face_mask, [True, True])

    return xr.Dataset(
        data_vars={
            "face_mask": xr.DataArray(face_mask, dims=dimensions[ArakawaCGridKind.face]),
            "back_mask": xr.DataArray(back_mask, dims=dimensions[ArakawaCGridKind.back]),
            "left_mask": xr.DataArray(left_mask, dims=dimensions[ArakawaCGridKind.left]),
            "node_mask": xr.DataArray(node_mask, dims=dimensions[ArakawaCGridKind.node]),
        },
        coords=coords,
    )



"""
:mod:`emsarray` supports multiple data conventions.
Each supported convention :ref:`implements the Convention interface <interface>`
to provide a common base set of functionality.

Calling :func:`.get_dataset_convention` with a dataset will inspect the dataset
and return the most appropriate :class:`Convention` implementation to use.
This can also be accessed using the :ref:`accessor`.

In some cases, the automatic :ref:`accessor`
will not be able to guess the file convention.
In other cases, you might want to override some default parameters.
Convention instances can be instantiated directly in this case.
Refer to each Convention implementation for details.
"""
from ._base import Convention, GridKind, Index, SpatialIndexItem, Specificity
from ._registry import get_dataset_convention, register_convention
from ._utils import open_dataset
from .arakawa_c import ArakawaC
from .grid import CFGrid1D, CFGrid2D
from .shoc import ShocSimple, ShocStandard
from .ugrid import UGrid

__all__ = [
    "Convention", "GridKind", "Index", "SpatialIndexItem", "Specificity",
    "get_dataset_convention", "register_convention",
    "open_dataset",
    "ArakawaC",
    "CFGrid1D", "CFGrid2D",
    "ShocSimple", "ShocStandard",
    "UGrid",
]



from __future__ import annotations

import logging
import sys
import warnings
from contextlib import suppress
from functools import cached_property
from itertools import chain
from typing import Iterable, List, Optional, Tuple, Type

import xarray as xr

from ._base import Convention

if sys.version_info >= (3, 10):
    from importlib import metadata
else:
    import importlib_metadata as metadata


logger = logging.getLogger(__name__)


class ConventionRegistry:
    registered_conventions: List[Type[Convention]]

    def __init__(self) -> None:
        self.registered_conventions = []

    @cached_property
    def conventions(self) -> Iterable[Type[Convention]]:
        """
        A list of all the registered Convention subclasses.
        This includes those registered via entry points
        and those registered via :func:`register_convention`.

        Returns
        -------
        list of Convention subclasses
            All the registered :class:`~emsarray.conventions.Convention` subclasses

        See also
        --------
        :func:`entry_point_conventions`
        :func:`register_convention`
        """
        conventions = []
        seen = set()
        # Construct a list of all registered conventions in the order that they
        # were registered. Manually registered conventions are prioritised.
        # Duplicates are removed.
        for convention in chain(self.registered_conventions, self.entry_point_conventions):
            if convention not in seen:
                conventions.append(convention)
                seen.add(convention)
        return conventions

    @cached_property
    def entry_point_conventions(self) -> List[Type[Convention]]:
        """
        Find all conventions registered via the ``emsarray.conventions`` entry point.
        This list is cached.

        Returns
        -------
        list of Convention subclasses
            All the :class:`~emsarray.conventions.Convention` subclasses registered
            via the ``emsarray.conventions`` entry point.
        """
        return list(entry_point_conventions())

    def add_convention(self, convention: Type[Convention]) -> None:
        """Register a Convention subclass with this registry.
        Datasets will be checked against this Convention when guessing file types.
        """
        with suppress(AttributeError):
            del self.conventions
        self.registered_conventions.append(convention)

    def match_conventions(self, dataset: xr.Dataset) -> List[Tuple[Type[Convention], int]]:
        """
        Get all :class:`~.Convention` implementations that support this dataset.

        Parameters
        ----------
        dataset : :class:`xarray.Dataset`
            The dataset to inspect

        Returns
        -------
        list of matches
            A list of ``(Convention class, match specificity)`` tuples.
            A higher specificity means a better match.
            The list of matches will be ordered from most to least specific.
        """
        matches: List[Tuple[Type[Convention], int]] = []
        for convention in self.conventions:
            match = convention.check_dataset(dataset)
            if match is not None:
                matches.append((convention, match))
        return sorted(matches, key=lambda m: m[1], reverse=True)

    def guess_convention(self, dataset: xr.Dataset) -> Optional[Type[Convention]]:
        """
        Guess the correct :class:`.Convention` implementation for a dataset.
        """
        matches = self.match_conventions(dataset)
        if matches:
            # Return the best match
            return matches[0][0]
        else:
            return None


registry = ConventionRegistry()


def get_dataset_convention(dataset: xr.Dataset) -> Optional[Type[Convention]]:
    """Find the most appropriate Convention subclass for this dataset.

    Parameters
    ----------
    dataset
        The dataset to introspect

    Returns
    -------
    :class:`.Convention`
        A :class:`.Convention` subclass appropriate for this dataset,
        or None if nothing appropriate can be found.

    Example
    -------

    .. code-block:: python

        >>> ds = emsarray.tutorial.open_dataset('austen.nc')
        >>> get_dataset_convention(ds)
        emsarray.conventions.ugrid.UGrid
    """
    return registry.guess_convention(dataset)


def entry_point_conventions() -> Iterable[Type[Convention]]:
    """
    Finds conventions registered using entry points
    """
    seen = set()

    groups = [
        ('emsarray.conventions', False),
        ('emsarray.formats', True),
    ]
    for group, deprecated in groups:
        for entry_point in metadata.entry_points(group=group):
            if deprecated:
                warnings.warn(
                    '`emsarray.formats` entrypoint has been renamed to `emsarray.conventions`. '
                    f'Update `{entry_point.name} = {entry_point.value}` to use the new entrypoint name.',
                    category=DeprecationWarning)
            try:
                obj = entry_point.load()
            except (AttributeError, ImportError):
                logger.exception("Error loading entry point %s", entry_point)
                continue

            if not (isinstance(obj, type) and issubclass(obj, Convention)):
                logger.error(
                    "Entry point `%s = %s` refers to %r not a Convention subclass",
                    entry_point.name, entry_point.value, obj)
                continue

            if obj not in seen:
                yield obj
                seen.add(obj)


def register_convention(convention: Type[Convention]) -> Type[Convention]:
    """
    Register a Convention subclass, used for guessing file types.
    Can be used as a decorator.

    This function is useful for making convention classes for internal project use.
    If you are distributing an emsarray convention class as a Python package,
    see :func:`entry_point_conventions` for registering conventions using an ``entry_point``.

    Example
    -------
    >>> from emsarray import conventions
    >>> @conventions.register
    ... class FooConvention(Convention):
    ...     pass

    See also
    --------
    :func:`entry_point_conventions`
    """
    registry.add_convention(convention)
    return convention



import argparse
import logging
from pathlib import Path

import pandas as pd

import emsarray
from emsarray.cli import BaseCommand, CommandException
from emsarray.operations import point_extraction
from emsarray.utils import to_netcdf_with_fixes

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Extract points from a dataset"
    description = """
    Extract points in a CSV file from a dataset
    and save this point file to a new dataset.
    All columns in the CSV are merged in to the new point dataset.
    """

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "input_path", type=Path,
            metavar='input-dataset',
            help="Path to input netCDF4 file")
        parser.add_argument(
            'points', type=Path, metavar='points',
            help=(
                "Path to a CSV file with the points to extract"
            ))
        parser.add_argument(
            "output_path", type=Path,
            metavar='output-dataset',
            help="Path to output netCDF4 file")

        parser.add_argument(
            "-c", "--coordinate-columns", type=str, nargs=2,
            metavar=('LON', 'LAT'),
            default=("lon", "lat"),
            help=(
                "Names of the longitude and latitude columns in the CSV file. "
                "Defaults to 'lon' and 'lat'."
            ))

        parser.add_argument(
            "-d", "--point-dimension", type=str,
            metavar=('DIM'),
            default="point",
            help=(
                "Name of the new dimension to index the point data"
            ))

    def handle(self, options: argparse.Namespace) -> None:
        logger.info("Extracting points from %r", str(options.input_path))
        dataset = emsarray.open_dataset(options.input_path)
        dataframe = pd.read_csv(options.points)

        try:
            point_data = point_extraction.extract_dataframe(
                dataset, dataframe, options.coordinate_columns,
                point_dimension=options.point_dimension)
        except point_extraction.NonIntersectingPoints as err:
            rows = dataframe.iloc[err.indices]
            raise CommandException(
                f"Error extracting points: the points in the following rows "
                f"did not intersect the dataset geometry:\n"
                f"{rows.head()}\n"
                f"(total rows: {len(rows)})")
        try:
            time_name = dataset.ems.get_time_name()
        except KeyError:
            time_name = None

        to_netcdf_with_fixes(
            point_data, options.output_path, time_variable=time_name)



from __future__ import annotations

import abc
import dataclasses
import enum
import logging
from functools import cached_property
from typing import (
    TYPE_CHECKING, Any, Callable, Dict, FrozenSet, Generic, Hashable, List,
    Optional, Tuple, TypeVar, Union, cast
)

import numpy as np
import xarray as xr
from shapely.geometry import Point, Polygon
from shapely.geometry.base import BaseGeometry

from emsarray import utils
from emsarray.compat.shapely import SpatialIndex
from emsarray.operations import depth
from emsarray.plot import (
    _requires_plot, animate_on_figure, plot_on_figure,
    polygons_to_patch_collection
)
from emsarray.state import State
from emsarray.types import Pathish

if TYPE_CHECKING:
    # Import these optional dependencies only during type checking
    from cartopy.crs import CRS
    from matplotlib.animation import FuncAnimation
    from matplotlib.axes import Axes
    from matplotlib.collections import PatchCollection
    from matplotlib.figure import Figure
    from matplotlib.quiver import Quiver

logger = logging.getLogger(__name__)


DataArrayOrName = Union[Hashable, xr.DataArray]

#: Some type that can enumerate the different :ref:`grid types <grids>`
#: present in a dataset.
#: This can be an :class:`enum.Enum` listing each different kind of grid.
#:
#: :data:`Index` values will be included in the feature properties
#: of exported geometry from :mod:`emsarray.operations.geometry`.
#: If the index type includes the grid kind,
#: the grid kind needs to be JSON serializable.
#: The easiest way to achieve this is to make your GridKind type subclass :class:`str`:
#:
#: .. code-block:: python
#:
#:     class MyGridKind(str, enum.Enum):
#:         face = 'face'
#:         edge = 'edge'
#:         node = 'node'
#:
#: For cases where the convention only supports a single grid,
#: a singleton enum can be used.
#:
#: More esoteric cases involving datasets with a potentially unbounded numbers of grids
#: can use a type that supports this instead.
GridKind = TypeVar("GridKind")

#: An :ref:`index <indexing>` to a specific point on a grid in this convention.
#: For conventions with :ref:`multiple grids <grids>` (e.g. cells, edges, and nodes),
#: this should be a tuple whos first element is :data:`.GridKind`.
#: For conventions with a single grid, :data:`.GridKind` is not required.
Index = TypeVar("Index")


@dataclasses.dataclass
class SpatialIndexItem(Generic[Index]):
    """Information about an item in the STRtree spatial index for a dataset.

    See also
    --------
    :attr:`.Convention.spatial_index`
    """

    #: The linear index of this cell
    linear_index: int

    #: The native index of this cell
    index: Index

    #: The geographic shape of this cell
    polygon: Polygon

    def __repr__(self) -> str:
        items = {
            'index': f'{self.index}/{self.linear_index}',
            'polygon': self.polygon.wkt,
        }
        item_str = ' '.join(f'{key}: {value}' for key, value in items.items())
        return f'<{type(self).__name__} {item_str}>'

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, SpatialIndexItem):
            return NotImplemented

        # SpatialIndexItems are only for cells / polygons, so we only need to
        # compare the linear indices. The polygon attribute is not orderable,
        # so comparing on that is going to be unpleasant.
        return self.linear_index < other.linear_index


class Specificity(enum.IntEnum):
    """
    How specific a match is when autodetecting a convention.
    Matches with higher specificity will be prioritised.

    General conventions such as CF Grid are low specificity,
    as many conventions extend and build on CF Grid conventions.

    The SHOC conventions extend the CF grid conventions,
    so a SHOC file will be detected as both CF Grid and SHOC.
    :class:`.ShocStandard` should return a higher specificity
    so that the correct convention implementation is used.
    """
    LOW = 10
    MEDIUM = 20
    HIGH = 30


class Convention(abc.ABC, Generic[GridKind, Index]):
    """
    Each supported geometry convention represents data differently.
    The :class:`Convention` class abstracts these differences away,
    allowing developers to query and manipulate datasets
    without worrying about the details.
    See :ref:`Supported dataset conventions <supported_conventions>`
    for a list of implemented conventions.

    All conventions have the concept of a cell at a geographic location,
    vertically stacked layers of cells,
    and multiple timesteps of data.
    A convention may support additional grids, such as face edges and vertices.
    Refer to :ref:`grids` for more information.

    A cell can be addressed using a linear index or a native index.
    A linear index is always an :class:`int`,
    while the native index type will depend on the specific convention.
    You can convert between a linear and a native index
    using :meth:`.ravel_index` and :meth:`.unravel_index`.
    Refer to :ref:`indexing` for more information.

    The depths of each layer can be found using :meth:`.get_depths`.
    The timesteps in a dataset can be found using :meth:`.get_times`.
    """
    #: The :class:`xarray.Dataset` instance for this :class:`Convention`
    dataset: xr.Dataset

    def __init__(self, dataset: xr.Dataset):
        """
        Make a new convention instance for this dataset.
        """
        self.check_validity(dataset)
        self.dataset = dataset

    @classmethod
    def check_validity(cls, dataset: xr.Dataset) -> None:
        """Checks that the dataset is OK to use.
        Called during __init__, and raises exceptions if the dataset has problems.
        """
        pass  # Subclasses can override this. By default, no checks are made

    @classmethod
    @abc.abstractmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        """
        Check if a dataset uses this convention.

        This may check for variables of the correct dimensions,
        the presence of specific attributes,
        or the layout of the dataset dimensions.

        Specific subclasses must implement this function.
        It will be called by the convention autodetector
        when guessing the correct convention for a dataset.

        If the dataset matches, the return value indicates how specific the match is.
        When autodetecting the correct convention implementation
        the convention with the highest specicifity will be used.
        Many conventions extend the CF grid conventions,
        so the CF Grid convention classes will match many datasets.
        However this match is very generic.
        A more specific implementation such as SHOC may be supported.
        The SHOC convention implementation should return a higher specicifity than the CF grid convention.

        Parameters
        ----------
        dataset : :class:`xarray.Dataset`
            The dataset instance to inspect.

        Returns
        -------
        int, optional
            If this convention implementation can handle this dataset
            some integer greater than zero is returned.
            The higher the number, the more specific the support.
            If the dataset does not match this convention, None is returned.
            Values on the :class:`~emsarray.conventions.Specificity` enum
            are used by :mod:`emsarray` itself to indicated specificity.
            New convention implementations are free to use these values,
            or use any integer value.

        Example
        -------
        >>> import xarray as xr
        ... from emsarray.conventions.shoc import ShocStandard
        ... from emsarray.conventions.ugrid import UGrid
        ... dataset = xr.open_dataset("./tests/datasets/shoc_standard.nc")
        >>> ShocStandard.check_dataset(dataset)
        True
        >>> UGrid.check_dataset(dataset)
        False
        """
        pass

    def bind(self) -> None:
        """
        Bind this :class:`.Convention` instance as the default convention
        for the :class:`xarray.Dataset`.
        This convention instance will be assigned to :attr:`dataset.ems`.

        You can use a Convention instance without binding it to a Dataset,
        binding is only necessary if you need to use the :attr:`dataset.ems` accessor.

        .. note::

            If you use :func:`emsarray.open_dataset` or :attr:`dataset.ems`
            to autodetect the dataset convention you do not need to call this method.
            :meth:`Convention.bind` is only useful if you manually construct a :class:`Convention`.

        Example
        -------

        .. code-block:: python

            # Open a dataset built using the GRASS convention
            dataset = xarray.open_dataset("grass-dataset.nc")

            # Construct a Grass instance for the dataset and bind it
            convention = Grass(dataset)
            convention.bind()

            # dataset.ems is now the bound convention
            assert dataset.ems is convention

        If the dataset already has a bound convention, an error is raised.
        To bind a new convention to a dataset, make a copy of the dataset first:

        .. code-block:: python

            new_dataset = dataset.copy()
            convention = Grass(new_dataset)
            convention.bind()
        """
        state = State.get(self.dataset)
        if state.is_bound():
            raise ValueError(
                "A convention has already been bound to this dataset, "
                "cannot assign a new convention.")
        state.bind_convention(self)

    def _get_data_array(self, data_array: DataArrayOrName) -> xr.DataArray:
        """
        Utility to help get a data array for this dataset.
        If a string is passed in, the matching data array is fetched from the dataset.
        If a data array is passed in, it is inspected to ensure the dimensions match
        before being returned as-is.

        This is useful for methods that support being passed either
        the name of a data array or a data array instance.
        """
        if isinstance(data_array, xr.DataArray):
            utils.check_data_array_dimensions_match(self.dataset, data_array)
            return data_array
        else:
            return self.dataset[data_array]

    def get_time_name(self) -> Hashable:
        """Get the name of the time variable in this dataset."""
        for name, variable in self.dataset.variables.items():
            if variable.attrs.get('standard_name') == 'time':
                return name
        raise KeyError("Dataset does not have a time dimension")

    def get_depth_name(self) -> Hashable:
        """Get the name of the layer depth coordinate variable.
        For datasets with multiple depth variables, this should be the one that
        represents the centre of the layer, not the bounds.

        Note that this is the name of the coordinate variable,
        not the name of the dimension, for datasets where these differ.
        """
        return self.get_all_depth_names()[0]

    def get_all_depth_names(self) -> List[Hashable]:
        """Get the names of all depth layers.
        Some datasets include both a depth layer centre,
        and the depth layer 'edges'.

        Note that this is the names of the coordinate variables,
        not the names of the dimensions, for datasets where these differ.
        """
        depth_names = []
        for name in self.dataset.variables.keys():
            data_array = self.dataset[name]

            if not (
                data_array.attrs.get('axis') == 'Z'
                or data_array.attrs.get('cartesian_axis') == 'Z'
                or data_array.attrs.get('coordinate_type') == 'Z'
                or data_array.attrs.get('standard_name') == 'depth'
            ):
                continue

            try:
                # If the variable is defined on a grid,
                # it is more likely to be a bathymetry variable
                # not the coordinate for the depth layers.
                self.get_grid_kind_and_size(data_array)
                continue
            except ValueError:
                # The variable isn't on a grid - this is good!
                pass

            depth_names.append(name)

        return depth_names

    def get_depths(self) -> np.ndarray:
        """Get the depth of each vertical layer in this dataset.

        Returns
        -------
        :class:`numpy.ndarray`
            An array of depths, one per vertical layer in the dataset.
        """
        return cast(np.ndarray, self.dataset.variables[self.get_depth_name()].values)

    def get_times(self) -> np.ndarray:
        """Get all timesteps in this dataset.

        Returns
        -------
        :class:`numpy.ndarray`
            An array of datetimes.
            The datetimes will be whatever native format the dataset uses,
            likely :class:`numpy.datetime64`.
        """
        return cast(np.ndarray, self.dataset.variables[self.get_time_name()].values)

    @abc.abstractmethod
    def ravel_index(self, index: Index) -> int:
        """Convert a convention native index to a linear index.

        Each conventnion has a different native index type,
        read the specific conventnion documentation for more information.

        Parameters
        ----------
        index : :data:`.Index`
            The native index to ravel.

        Returns
        -------
        :class:`int`
            The linear index for that native index.

        Example
        -------

        If the dataset used the CF Grid conventions,
        across a (latitude, longitude) grid of size (30, 40):

        .. code-block:: python

            >>> temp = dataset.data_vars['temp']
            >>> temp.dims
            ('t', 'z', 'y', 'x')
            >>> temp.shape
            (10, 20, 30, 40)
            >>> dataset.ems.ravel_index((3, 4))
            124

        Cell polygons are indexed in the same order as the linear indices for cells.
        To find the polygon for the cell with the native index ``(3, 4)``:

        .. code-block:: python

            >>> index = (3, 4)
            >>> linear_index = dataset.ems.ravel_index(index)
            >>> polygon = dataset.ems.polygons[linear_index]

        See also
        --------
        :meth:`.unravel_index`
        """
        pass

    @abc.abstractmethod
    def unravel_index(
        self,
        linear_index: int,
        grid_kind: Optional[GridKind] = None,
    ) -> Index:
        """Convert a linear index to a conventnion native index.

        Each conventnion has a different native index type,
        read the specific conventnion documentation for more information.

        Parameters
        ----------
        linear_index : int
            The linear index to unravel.
        grid_kind : GridKind, optional
            Used to indicate what kind of index is being unravelled,
            for conventions with multiple grids.
            Optional, if not provided this will return the unravelled face index.

        Returns
        -------
        :data:`.Index`
            The convention native index for that same cell

        Example
        -------

        If the dataset used the CF Grid conventions,
        across a (latitude, longitude) grid of size (30, 40):

        .. code-block:: python

            >>> temp = dataset.data_vars['temp']
            >>> temp.dims
            ('t', 'z', 'y', 'x')
            >>> temp.shape
            (10, 20, 30, 40)
            >>> dataset.ems.unravel_index(124)
            (3, 4)

        See also
        --------
        :meth:`.ravel_index`
        """
        pass

    @property
    @abc.abstractmethod
    def grid_kinds(self) -> FrozenSet[GridKind]:
        """
        All of the :data:`grid kinds <GridKind>` this dataset includes.
        """
        pass

    @property
    @abc.abstractmethod
    def default_grid_kind(self) -> GridKind:
        """
        The default :data:`grid kind <GridKind>` for this dataset.
        For most datasets this should be the face grid.
        """
        pass

    @abc.abstractmethod
    def get_grid_kind_and_size(
        self, data_array: xr.DataArray,
    ) -> Tuple[GridKind, int]:
        """
        Determines the relevant index kind and the extent of the linear index space
        for this data array.

        If the data array is not indexable using the native index types
        a ValueError is raised.

        Parameters
        ----------
        data_array
            The data array to introspect

        Returns
        -------
        tuple of :data:`GridKind` and int

        Raises
        ------
        `ValueError`
            If the data array passed in is not indexable using any native index type
            a ValueError is raised.
            Depth coordinates or time coordinates are examples of data arrays
            that will not be indexable and will raise an error.

        Example
        -------
        For a :class:`UGRID <.ugrid.UGrid>` dataset
        with temperature data defined at the cell centres
        and current defined as flux through the cell edges:

        .. code-block:: python

            >>> dataset.ems.topology.face_count
            4
            >>> dataset.ems.topology.edge_count
            9
            >>> dataset.data_vars['temp'].dims
            ('time', 'depth', 'face')
            >>> dataset.data_vars['u1'].dims
            ('time', 'depth', 'edge')
            >>> dataset.ems.get_grid_kind_and_size(dataset.data_vars['temp'])
            (UGridKind.face, 4)
            >>> dataset.ems.get_grid_kind_and_size(dataset.data_vars['u1'])
            (UGridKind.edge, 9)
        """
        pass

    @abc.abstractmethod
    def make_linear(self, data_array: xr.DataArray) -> xr.DataArray:
        """
        Flatten the surface dimensions of a :class:`~xarray.DataArray`,
        returning a flatter :class:`np.ndarray` indexed in the same order as the linear index.

        For DataArrays with extra dimensions such as time or depth,
        only the surface dimensions are flattened.
        Other dimensions are left as is.

        For datasets with multiple grids,
        with data defined on edges or vertices for example,
        this will flatten those data arrays in the correct linear order
        to be indexed by the relevant index type.

        Parameters
        ----------
        data_array
            One of the data variables from this dataset.

        Returns
        -------
        :class:`xarray.DataArray`
            A new data array, where all the surface dimensions have been flattened in to one linear array.
            The values for each cell, in the same order as the linear index for this dataset.
            Any other dimensions, such as depth or time, will be retained.
        """
        pass

    @cached_property  # type: ignore
    @_requires_plot
    def data_crs(self) -> CRS:
        """
        The coordinate reference system that coordinates in this dataset are
        defined in.
        Used by :meth:`.make_patch_collection` and :meth:`.make_quiver`.
        Defaults to :class:`cartopy.crs.PlateCarree`.
        """
        # Lazily imported here as cartopy is an optional dependency
        from cartopy.crs import PlateCarree
        return PlateCarree()

    @_requires_plot
    def plot_on_figure(
        self,
        figure: Figure,
        scalar: Optional[DataArrayOrName] = None,
        vector: Optional[Tuple[DataArrayOrName, DataArrayOrName]] = None,
        title: Optional[str] = None,
    ) -> None:
        """Plot values for a :class:`~xarray.DataArray`
        on a :mod:`matplotlib` :class:`~matplotlib.figure.Figure`.

        The data array can either be passed in directly,
        or the name of a data array on this :attr:`Convention.dataset` instance.
        The data array does not have to come from the same dataset,
        as long as the dimensions are the same.

        Parameters
        ----------
        figure
            The :class:`~matplotlib.figure.Figure` instance to plot this on.
        scalar : data array
            The :class:`~xarray.DataArray` to plot,
            or the name of an existing DataArray in this Dataset.
            This method will only plot a single time step and depth layer.
            Callers are responsible for selecting a single slice.
        vector : tuple of data arrays
            A tuple of the *u* and *v* components of a vector.
            The components should be a :class:`~xarray.DataArray`,
            or the name of an existing DataArray in this Dataset.
            This method will only plot a single time step and depth layer.
            Callers are responsible for selecting a single slice.

        See also
        --------
        :func:`.plot.plot_on_figure` : The underlying implementation
        """
        kwargs: Dict[str, Any] = {}
        if scalar is not None:
            kwargs['scalar'] = self._get_data_array(scalar)

        if vector is not None:
            kwargs['vector'] = tuple(map(self._get_data_array, vector))

        if title is not None:
            kwargs['title'] = title
        elif scalar is not None and vector is None:
            # Make a title out of the scalar variable, but only if a title
            # hasn't been supplied and we don't also have vectors to plot.
            #
            # We can't make a good name from vectors,
            # as they are in two variables with names like
            # 'u component of current' and 'v component of current'.
            #
            # Users can supply their own titles
            # if this automatic behaviour is insufficient
            kwargs['title'] = kwargs['scalar'].attrs.get('long_name')
            try:
                time_coordinate = self.dataset.variables[self.get_time_name()]
            except KeyError:
                pass
            else:
                # Add a time stamp when the time coordinate has a single value.
                # This happens when you `.sel()` a single time slice to plot -
                # as long as the time coordinate is a proper coordinate with
                # matching dimension name, not an auxiliary coordinate.
                if time_coordinate.size == 1:
                    time = time_coordinate.values
                    kwargs['title'] = kwargs['title'] + '\n' + str(time)

        plot_on_figure(figure, self, **kwargs)

    @_requires_plot
    def plot(
        self,
        scalar: Optional[DataArrayOrName] = None,
        vector: Optional[Tuple[DataArrayOrName, DataArrayOrName]] = None,
        title: Optional[str] = None,
    ) -> None:
        """Plot a data array and automatically display it.

        This method is most useful when working in Jupyter notebooks
        which display figures automatically.

        See also
        --------
        :meth:`.plot_on_figure`
        """
        from matplotlib import pyplot as plt
        self.plot_on_figure(plt.figure(), scalar=scalar, vector=vector, title=title)
        plt.show()

    @_requires_plot
    def animate_on_figure(
        self,
        figure: Figure,
        scalar: Optional[DataArrayOrName] = None,
        vector: Optional[Tuple[DataArrayOrName, DataArrayOrName]] = None,
        coordinate: Optional[DataArrayOrName] = None,
        title: Optional[Union[str, Callable[[Any], str]]] = None,
        **kwargs: Any,
    ) -> FuncAnimation:
        """
        Make an animated plot of a data array.

        For real world examples, refer to the ``examples/animation.ipynb`` notebook.

        Parameters
        ----------
        figure
            The :class:`matplotlib.figure.Figure` to plot the animation on
        data_array
            The :class:`xarray.DataArray` to plot.
            If a string is passed in,
            the variable with that name is taken from :attr:`dataset`.
        coordinate
            The coordinate to vary across the animation.
            Pass in either the name of a coordinate variable
            or coordinate variable itself.
            Optional, if not supplied the time coordinate
            from :meth:`get_time_name` is used.
            Other appropriate coordinates to animate over include depth.

        Returns
        -------
        :class:`matplotlib.animation.Animation`
            The data variable plotted as an animation on the figure.
            Call :meth:`Animation.save() <matplotlib.animation.Animation.save>`
            to save animation to a file, or display it in a Notebook using
            :meth:`Animation.to_jshtml() <matplotlib.animation.Animation.to_jshtml>`.

        See also
        --------
        :func:`.plot.animate_on_figure`
        """

        if coordinate is None:
            # Assume the user wants to plot along the time axis by default.
            coordinate = self.get_time_name()

        coordinate = self._get_data_array(coordinate)

        if len(coordinate.dims) != 1:
            raise ValueError("Coordinate variable must be one dimensional")

        coordinate_dim = coordinate.dims[0]

        if scalar is not None:
            scalar = self._get_data_array(scalar)
            if coordinate_dim not in scalar.dims:
                raise ValueError("Scalar dimensions do not match coordinate axis to animate along")
            kwargs['scalar'] = scalar

        if vector is not None:
            vector = (self._get_data_array(vector[0]), self._get_data_array(vector[1]))
            if not all(coordinate_dim in component.dims for component in vector):
                raise ValueError("Vector dimensions do not match coordinate axis to animate along")
            kwargs['vector'] = vector

        if title is not None:
            kwargs['title'] = title
        elif scalar is not None and vector is None:
            # Make a title out of the scalar variable, but only if a title
            # hasn't been supplied and we don't also have vectors to plot.
            #
            # We can't make a good name from vectors,
            # as they are in two variables with names like
            # 'u component of current' and 'v component of current'.
            #
            # Users can supply their own titles
            # if this automatic behaviour is insufficient
            title_bits = []
            if 'long_name' in scalar.attrs:
                title_bits.append(str(scalar.attrs['long_name']))
            elif scalar.name:
                title_bits.append(str(scalar.name))

            if 'long_name' in coordinate.attrs:
                title_bits.append(str(coordinate.attrs['long_name']) + ': {}')
            elif coordinate.name:
                title_bits.append(str(coordinate.name) + ': {}')
            else:
                title_bits.append('{}')
            kwargs['title'] = '\n'.join(title_bits)

        return animate_on_figure(figure, self, coordinate=coordinate, **kwargs)

    @_requires_plot
    def make_patch_collection(
        self,
        data_array: Optional[DataArrayOrName] = None,
        **kwargs: Any,
    ) -> PatchCollection:
        """
        Make a :class:`~matplotlib.collections.PatchCollection`
        from the geometry of this :class:`~xarray.Dataset`.
        This can be used to make custom matplotlib plots from your data.

        If a :class:`~xarray.DataArray` is passed in,
        the values of that are assigned to the PatchCollection `array` parameter.

        Parameters
        ----------
        data_array : Hashable or :class:`xarray.DataArray`, optional
            A data array, or the name of a data variable in this dataset. Optional.
            If given, the data array is :meth:`linearised <.make_linear>`
            and passed to :meth:`PatchCollection.set_array() <matplotlib.cm.ScalarMappable.set_array>`.
            The data is used to colour the patches.
            Refer to the matplotlib documentation for more information on styling.
        **kwargs
            Any keyword arguments are passed to the
            :class:`~matplotlib.collections.PatchCollection` constructor.

        Returns
        -------
        :class:`~matplotlib.collections.PatchCollection`
            A PatchCollection constructed using the geometry of this dataset.

        Example
        -------

        .. code-block:: python

            import cartopy.crs as ccrs
            import matplotlib.pyplot as plt
            import emsarray

            figure = plt.figure(figsize=(10, 8))
            axes = plt.subplot(projection=ccrs.PlateCarree())
            axes.set_aspect(aspect='equal', adjustable='datalim')

            ds = emsarray.open_dataset("./tests/datasets/ugrid_mesh2d.nc")
            ds = ds.isel(record=0, Mesh2_layers=-1)
            patches = ds.ems.make_patch_collection('temp')
            axes.add_collection(patches)
            figure.colorbar(patches, ax=axes, location='right', label='meters')

            axes.set_title("Depth")
            axes.autoscale()
            figure.show()
        """
        if data_array is not None:
            if 'array' in kwargs:
                raise TypeError(
                    "Can not pass both `data_array` and `array` to make_patch_collection"
                )

            data_array = self._get_data_array(data_array)

            data_array = self.make_linear(data_array)
            if len(data_array.dims) > 1:
                raise ValueError(
                    "Data array has too many dimensions - did you forget to "
                    "select a single timestep or a single depth layer?")

            values = data_array.values[self.mask]
            kwargs['array'] = values
            if 'clim' not in kwargs:
                kwargs['clim'] = (np.nanmin(values), np.nanmax(values))

        if 'transform' not in kwargs:
            kwargs['transform'] = self.data_crs

        return polygons_to_patch_collection(self.polygons[self.mask], **kwargs)

    @_requires_plot
    def make_quiver(
        self,
        axes: Axes,
        u: Optional[DataArrayOrName] = None,
        v: Optional[DataArrayOrName] = None,
        **kwargs: Any,
    ) -> Quiver:
        from matplotlib.quiver import Quiver

        x, y = np.transpose(self.face_centres)

        # A Quiver needs some values when being initialized.
        # We don't always want to provide values to the quiver,
        # sometimes preferring to fill them in later,
        # so `u` and `v` are optional.
        # If they are not provided, we set default quiver values of `np.nan`.
        values: Union[Tuple[np.ndarray, np.ndarray], Tuple[float, float]]
        values = np.nan, np.nan

        if u is not None and v is not None:
            u, v = self._get_data_array(u), self._get_data_array(v)

            if u.dims != v.dims:
                raise ValueError(
                    "Vector data array dimensions must be identical!\n"
                    f"u dimensions: {tuple(u.dims)}\n"
                    f"v dimensions: {tuple(v.dims)}"
                )

            u, v = self.make_linear(u), self.make_linear(v)

            if len(u.dims) > 1:
                raise ValueError(
                    "Vector data arrays have too many dimensions - did you forget to "
                    "select a single timestep or a single depth layer?")

            values = u.values, v.values

        if 'transform' not in kwargs:
            kwargs['transform'] = self.data_crs

        return Quiver(axes, x, y, *values, **kwargs)

    @property
    @abc.abstractmethod
    def polygons(self) -> np.ndarray:
        """A :class:`np.ndarray` of :class:`Polygon` instances representing the cells in this dataset.

        The order of the polygons in the list corresponds to the linear index of this dataset.
        Not all valid cell indices have a polygon, these holes are represented as :data:`None` in the list.
        If you want a list of just polygons, apply the :attr:`mask <Convention.mask>`:

        .. code-block:: python

            dataset = emsarray.open_dataset("...")
            only_polygons = dataset.ems.polygons[dataset.ems.mask]

        See also
        --------
        :meth:`ravel_index`
        :attr:`mask`
        """
        pass

    @cached_property
    def face_centres(self) -> np.ndarray:
        """
        A numpy :class:`~numpy.ndarray` of face centres, which are (x, y) pairs.
        The first dimension will be the same length and in the same order
        as :attr:`Convention.polygons`,
        while the second dimension will always be of size 2.
        """
        # This default implementation simply finds the centroid of each polygon.
        # Subclasses are free to override this if the particular convention and dataset
        # provides the cell centres as a data array.
        centres = np.array([
            polygon.centroid.coords[0] if polygon is not None else [np.nan, np.nan]
            for polygon in self.polygons
        ])
        return cast(np.ndarray, centres)

    @cached_property
    def mask(self) -> np.ndarray:
        """
        A boolean :class:`np.ndarray` indicating which cells have valid polygons.
        This can be used to select only items from linear arrays
        that have a corresponding polygon.

        .. code-block:: python

            dataset = emsarray.open_dataset("...")
            mask = dataset.ems.mask
            plottable_polygons = dataset.ems.polygons[mask]
            plottable_values = dataset.ems.make_linear("eta")[mask]

        See also
        --------
        :meth:`Convention.make_linear`
        """
        mask = np.fromiter(
            (p is not None for p in self.polygons),
            dtype=bool, count=self.polygons.size)
        return cast(np.ndarray, mask)

    @cached_property
    def spatial_index(self) -> SpatialIndex[SpatialIndexItem[Index]]:
        """
        A shapely :class:`strtree.STRtree` spatial index of all cells in this dataset.
        This allows for fast spatial lookups, querying which cells lie at
        a point, or which cells intersect a geometry.

        Querying the index with :meth:`strtree.STRtree.query_items` will return a list
        of :class:`SpatialIndexItem` instances. This will contain all cells
        which have envelopes overlapping the queried geometry. The caller must
        then refine the results further, by checking for intersection, cover,
        or contains for example.

        Example
        -------

        To find the indices of all cells that overlap a given shape:

        .. code-block:: python

            indices = [
                item.index
                for item, polygon in dataset.ems.spatial_index.query(shape)
                if polygon.intersects(shape)
            ]

        See also
        --------
        :class:`.SpatialIndexItem`
        """
        logger.info("Building spatial index...")
        with utils.PerfTimer() as timer:
            items = [
                (poly, SpatialIndexItem(index, self.unravel_index(index), poly))
                for index, poly in enumerate(self.polygons)
                if poly is not None
            ]
            spatial_index = SpatialIndex(items)
        logger.debug("Build spatial index in %f seconds", timer.elapsed)
        return spatial_index

    def get_index_for_point(
        self,
        point: Point,
    ) -> Optional[SpatialIndexItem[Index]]:
        """
        Find the index for a :class:`point <Point>` in the dataset.

        Parameters
        ----------
        point : :class:`Point`
            The geographic point to query

        Returns
        -------
        :class:`SpatialIndexItem`, optional
            The :class:`SpatialIndexItem` for the point queried.
            This indicates the polygon that intersected the point
            and the index of that polygon in the dataset.

            If the point does not intersect the dataset, None is returned.

        Notes
        -----
        In the case where the point intersects multiple cells
        the cell with the lowest linear index is returned.
        This can happen if the point is exactly one of the cell vertices,
        or falls on a cell edge,
        or if the geometry of the dataset contains overlapping polygons.
        """
        hits: List[SpatialIndexItem] = sorted(
            item
            for polygon, item in self.spatial_index.query(point)
            if polygon.intersects(point)
        )
        if len(hits) > 0:
            return hits[0]
        return None

    @abc.abstractmethod
    def selector_for_index(self, index: Index) -> Dict[Hashable, int]:
        """
        Convert a convention native index into a selector
        that can be passed to :meth:`Dataset.isel <xarray.Dataset.isel>`.

        Parameters
        ----------
        index : :data:`Index`
            A convention native index

        Returns
        -------
        selector
            A dict suitable for passing to :meth:`xarray.Dataset.isel`
            that will select values at this index.

        See also
        --------
        :meth:`.select_index`
        :meth:`.select_point`
        :ref:`indexing`
        """
        pass

    def select_index(
        self,
        index: Index,
    ) -> xr.Dataset:
        """
        Return a new dataset that contains values only from a single index.
        This is much like doing a :func:`xarray.Dataset.isel()` on an index,
        but works with convention native index types.

        An index is associated with a grid kind.
        The returned dataset will only contain variables that were defined on this grid,
        with the single indexed point selected.
        For example, if the index of a face is passed in,
        the returned dataset will not contain any variables defined on an edge.

        Parameters
        ----------
        index : :data:`Index`
            The index to select.
            The index must be for the default grid kind for this dataset.

        Returns
        -------
        :class:`xarray.Dataset`
            A new dataset that is subset to the one index.

        Notes
        -----

        The returned dataset will most likely not have sufficient coordinate data
        to be used with a particular :class:`Convention` any more.
        The ``dataset.ems`` accessor will raise an error if accessed on the new dataset.
        """
        selector = self.selector_for_index(index)

        # Make a new dataset consisting of only data arrays that use at least
        # one of these dimensions.
        dims = set(selector.keys())
        names = [
            name for name, data_array in self.dataset.items()
            if dims.intersection(data_array.dims)
        ]
        dataset = utils.extract_vars(self.dataset, names)

        # Select just this point
        return dataset.isel(selector)

    def select_point(self, point: Point) -> xr.Dataset:
        """
        Return a new dataset that contains values for a single point.
        This is a shortcut for :meth:`get_index_for_point` and :meth:`select_index`.

        If the point is not inside the dataset a :class:`ValueError` is raised.

        Parameters
        ----------
        point : :class:`Point`
            The point to select

        Returns
        -------
        :class:`xarray.Dataset`
            A dataset of values at the point
        """
        index = self.get_index_for_point(point)
        if index is None:
            raise ValueError("Point did not intersect dataset")
        return self.select_index(index.index)

    @abc.abstractmethod
    def drop_geometry(self) -> xr.Dataset:
        """
        Return a new :class:`xarray.Dataset`
        with all geometry variables dropped.
        Useful when significantly transforming the dataset,
        such as :mod:`extracting point data <emsarray.operations.point_extraction>`.
        """
        pass

    @abc.abstractmethod
    def make_clip_mask(
        self,
        clip_geometry: BaseGeometry,
        *,
        buffer: int = 0,
    ) -> xr.Dataset:
        """
        Make a new Dataset that can be used to clip this dataset to only the
        cells that intersect some geometry.

        This dataset can be saved to a file to be reused to cut multiple
        datasets with identical shapes, such as a series of files representing
        multiple time series of a model.

        The mask can be applied to this dataset (or other datasets identical in
        shape) using :meth:`apply_clip_mask`.

        Parameters
        ----------
        clip_geometry : BaseGeometry
            The desired area to cut out. This can be any shapely geometry type,
            but will most likely be a polygon
        buffer : int, optional
            If set to a positive integer,
            a buffer of that many cells will be added around the clip region.
            This is useful if you need to clip to a particular area,
            but also would like to do some interpolation on the output cells.

        Returns
        -------
        :class:`xarray.Dataset`
            The mask

        See also
        --------
        :func:`apply_clip_mask`
        :func:`clip`
        """
        pass

    @abc.abstractmethod
    def apply_clip_mask(self, clip_mask: xr.Dataset, work_dir: Pathish) -> xr.Dataset:
        """
        Apply a clip mask to this dataset, and return a new dataset.
        Call :func:`make_clip_mask` to create a clip mask from a clip geometry.

        The ``clip_mask`` can be saved and loaded to disk if the mask needs to
        be reused across multiple datasets, such as multiple time series from
        one model.

        Depending on the implementation, the input dataset may be sliced in to
        multiple files during cutting, and the returned :class:`~xarray.Dataset`
        might be a multi-file :class:`~xarray.Dataset` built from these
        temporary files. The caller must either load the dataset in to memory
        using :meth:`~xarray.Dataset.load` or :meth:`~xarray.Dataset.compute`,
        or save the dataset to disk somewhere outside of the working directory
        before the working directory is cleaned up.

        Parameters
        ----------
        clip_mask : xarray.Dataset
            The mask, as made by :meth:`make_clip_mask`.
        work_dir : str or pathlib.Path
            A directory where temporary files can be written to.
            Callers must create and manage this temporary directory,
            perhaps using :obj:`tempfile.TemporaryDirectory`.

        Returns
        -------
        xarray.Dataset
            A new :class:`~xarray.Dataset` clipped using the mask
        """

    def clip(
        self,
        clip_geomery: BaseGeometry,
        work_dir: Pathish,
        *,
        buffer: int = 0,
    ) -> xr.Dataset:
        """
        Generates a clip mask and applies it in one step.

        See the documentation for :meth:`.make_clip_mask` and
        :meth:`.apply_clip_mask` for more details.

        Parameters
        ----------
        clip_geometry : BaseGeometry
            The desired area to cut out.
            This can be any shapely geometry type,
            but will most likely be a polygon
        work_dir : str or pathlib.Path
            A directory where temporary files can be written to.
            Callers must create and manage this temporary directory,
            perhaps using :obj:`tempfile.TemporaryDirectory`.
        buffer : int, optional
            If set to a positive integer,
            a buffer of that many cells will be added around the clip region.
            This is useful if you need to clip to a particular area,
            but also would like to do some interpolation on the output cells.

        Returns
        -------
        xarray.Dataset
            A new :class:`~xarray.Dataset` clipped using the mask
        """
        mask = self.make_clip_mask(clip_geomery, buffer=buffer)
        return self.apply_clip_mask(mask, work_dir=work_dir)

    def to_netcdf(self, path: Pathish, **kwargs: Any) -> None:
        """
        Save this dataset to a netCDF file, and also fix up the time units to
        make the EMS compatible.
        """
        try:
            time_variable = self.get_time_name()
        except KeyError:
            time_variable = None
        utils.to_netcdf_with_fixes(
            self.dataset, path, time_variable=time_variable, **kwargs)

    # Aliases for emsarray.operations

    def ocean_floor(self) -> xr.Dataset:
        """An alias for :func:`emsarray.operations.depth.ocean_floor`"""
        return depth.ocean_floor(
            self.dataset, self.get_all_depth_names(),
            non_spatial_variables=[self.get_time_name()])

    def normalize_depth_variables(
        self, positive_down: bool = True, deep_to_shallow: bool = True,
    ) -> xr.Dataset:
        """An alias for :func:`emsarray.operations.depth.normalize_depth_variables`"""
        return depth.normalize_depth_variables(
            self.dataset, self.get_all_depth_names(),
            positive_down=positive_down, deep_to_shallow=deep_to_shallow)



"""
SHOC (Sparse Hydrodynamic Ocean Code) is the hydrodynamic model used by the
`Coastal Environmental Modelling Team <https://research.csiro.au/cem/>`_
at `CSIRO <https://www.csiro.au/>`_.
There are two related conventions: SHOC standard and SHOC simple.

SHOC standard is an :mod:`Arakawa C grid <.arakawa_c>` dataset
with known coordinate names.
SHOC standard datasets can be simplified in to a SHOC simple dataset,
which is a :class:`curvilinear CF grid <.grid.CFGrid2D>`
with known coordinate names.

See also
--------
`SHOC documentation <https://research.csiro.au/cem/software/ems/hydro/strucutured-shoc/>`_
"""
from __future__ import annotations

import logging
from functools import cached_property
from typing import Hashable, List, Optional, Tuple

import xarray as xr

from ._base import Specificity
from .arakawa_c import ArakawaC, ArakawaCGridKind
from .grid import CFGrid2D, CFGrid2DTopology

logger = logging.getLogger(__name__)


class ShocStandard(ArakawaC):
    """
    SHOC datasets are :class:`.ArakawaC` datasets
    with predefined coordinate names for the four grids.
    """

    #: Coordinate names for SHOC datasets.
    #: The face coordinates are ``y_centre`` and ``x_centre``,
    #: left edge coordinates are ``y_left`` and ``x_left``,
    #: back edge coordinates are ``y_back`` and ``x_back``,
    #: and node coordinates are ``y_grid`` and ``x_grid``.
    coordinate_names = {
        ArakawaCGridKind.face: ('y_centre', 'x_centre'),
        ArakawaCGridKind.left: ('y_left', 'x_left'),
        ArakawaCGridKind.back: ('y_back', 'x_back'),
        ArakawaCGridKind.node: ('y_grid', 'x_grid'),
    }

    def get_depth_name(self) -> Hashable:
        return 'z_centre'

    def get_all_depth_names(self) -> List[Hashable]:
        return ['z_centre', 'z_grid']

    def get_time_name(self) -> Hashable:
        return 't'

    def drop_geometry(self) -> xr.Dataset:
        dataset = super().drop_geometry()
        dataset.attrs.pop('Conventions', None)
        return dataset


class ShocSimple(CFGrid2D):
    """
    SHOC standard datasets can be simplified down to SHOC simple datasets,
    which are :class:`~.grid.CFGrid2D` curvilinear grids.
    The latitude and longitude coordinate variables are named ``j`` and ``i``.
    Edge and node dimensions are dropped.
    """
    _dimensions: Tuple[Hashable, Hashable] = ('j', 'i')

    @cached_property
    def topology(self) -> CFGrid2DTopology:
        y_dimension, x_dimension = self._dimensions
        try:
            latitude = next(
                name for name, variable in self.dataset.variables.items()
                if variable.dims == self._dimensions
                and variable.attrs["standard_name"] == "latitude"
            )
            longitude = next(
                name for name, variable in self.dataset.variables.items()
                if variable.dims == self._dimensions
                and variable.attrs["standard_name"] == "longitude"
            )
        except StopIteration:
            raise ValueError("Could not find the necessary coordinate variables")

        return CFGrid2DTopology(self.dataset, latitude=latitude, longitude=longitude)

    @classmethod
    def check_dataset(cls, dataset: xr.Dataset) -> Optional[int]:
        if 'ems_version' not in dataset.attrs:
            return None
        if not set(dataset.dims).issuperset(cls._dimensions):
            return None
        return Specificity.HIGH

    def get_time_name(self) -> Hashable:
        return 'time'

    def get_depth_name(self) -> Hashable:
        return 'zc'

    def get_all_depth_names(self) -> List[Hashable]:
        return [self.get_depth_name()]



import numpy as np
import pytest
from shapely.geometry import Polygon

from emsarray.plot import polygon_to_patch, polygons_to_patch_collection


@pytest.mark.matplotlib
def test_polygon_to_patch():
    polygon = Polygon([(0, 0), (1, 0), (2, 2,), (0, 1), (0, 0)])
    patch = polygon_to_patch(polygon)
    for index, poly_coords in enumerate(polygon.exterior.coords):
        patch_coords = patch.get_xy()[index]
        assert tuple(poly_coords) == tuple(patch_coords)


@pytest.mark.matplotlib
def test_polygons_to_patch_collection():
    polygons = [
        Polygon([(i, 0), (i + 1, 0), (i + 1, 1), (i, 1), (i, 0)])
        for i in range(10)
    ]
    data = np.random.random(10) * 10
    patch_collection = polygons_to_patch_collection(
        polygons, array=data, cmap='autumn', clim=(0, 10))

    # Check that the polygons came through
    assert len(patch_collection.get_paths()) == 10

    # Check that keyword arguments were passed through
    assert patch_collection.get_cmap().name == 'autumn'
    np.testing.assert_equal(patch_collection.get_array(), data)



import warnings
from functools import wraps
from typing import Any, Optional, Type

import xarray as xr

from emsarray import conventions

from .conventions import Convention, GridKind, Index, get_dataset_convention


def _warn_old_new(old: str, new: str, **kwargs: Any) -> None:
    warnings.warn(
        f"{old} has been renamed to {new}",
        category=DeprecationWarning,
        **kwargs)


@wraps(get_dataset_convention)
def get_file_format(dataset: xr.Dataset, **kwargs: Any) -> Optional[Type[Convention]]:
    _warn_old_new(
        old="emsarray.formats.get_file_format",
        new="emsarray.conventions.get_dataset_convention",
        stacklevel=3)
    return get_dataset_convention(dataset, **kwargs)


class Format(Convention[GridKind, Index]):
    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        _warn_old_new(
            old="emsarray.formats.Format",
            new="emsarray.conventions.Convention",
            stacklevel=4)
        super().__init_subclass__(*args, **kwargs)


def __getattr__(name: str) -> Any:
    # This takes the place of `from emsarray.conventions import *`,
    # and will warn on any `from emsarray.formats import x, y, z` uses.
    if not name.startswith('__'):
        _warn_old_new(
            old=f'emsarray.formats.{name}',
            new=f'emsarray.conventions.{name}',
            stacklevel=3)
    return getattr(conventions, name)


# Even this module is deprecated! Warn about that too
_warn_old_new(old='emsarray.formats', new='emsarray.conventions', stacklevel=3)



import pathlib

import netCDF4

from emsarray import Convention

from .conventions.test_shoc_standard import make_dataset


def test_ems_accessor():
    dataset = make_dataset(j_size=10, i_size=20)
    assert isinstance(dataset.ems, Convention)


def test_to_netcdf(tmp_path: pathlib.Path):
    """
    Test that ``dataset.ems.to_netcdf()`` saves a netCDF4 file with some
    various fixes applied:

    * Time units corrected to be in a EMS-compatible format
    * _FillValue only set when needed

    This needs to be a proper ems dataset so that the accessor works.
    """
    xr_dataset = make_dataset(j_size=10, i_size=10)

    # There are a few problems we work around. Lets assert that they are still
    # problems before we check whether we fixed them.
    xr_dataset.to_netcdf(tmp_path / "bad.nc")
    with netCDF4.Dataset(tmp_path / "bad.nc", "r") as nc_dataset:
        # Formatted in a way that EMS can not parse
        time_units = nc_dataset.variables["t"].getncattr("units")
        assert time_units == "days since 1990-01-01T00:00:00+10:00"

        # The CF conventions state that coordinate variables should not have
        # needless _FillValues set, as all coordinates should be defined.
        # Some of our coordinates are not all defined - x_grid, y_grid for
        # datasets with holes being one example - but z_grid is fully defined
        # and should not have a _FillValue
        assert "_FillValue" in nc_dataset.variables["z_grid"].ncattrs()

    # Save it using dataset.ems.to_netcdf(...)
    xr_dataset.ems.to_netcdf(tmp_path / "good.nc")
    with netCDF4.Dataset(tmp_path / "good.nc", "r+") as nc_dataset:
        # The time units should now be in an EMS-compatible format
        time_units = nc_dataset.variables["t"].getncattr("units")
        assert time_units == 'days since 1990-01-01 00:00:00 +10:00'

        # _FillValue should not be set automatically for all variables
        assert "_FillValue" not in nc_dataset.variables["z_grid"].ncattrs()



"""
Operations for making a triangular mesh out of the polygons of a dataset.
"""
from typing import List, Tuple, cast

import xarray as xr
from shapely.geometry import LineString, MultiPoint, Polygon

Vertex = Tuple[float, float]
Triangle = Tuple[int, int, int]


def triangulate_dataset(
    dataset: xr.Dataset,
) -> Tuple[List[Vertex], List[Triangle], List[int]]:
    """
    Triangulate the polygon cells of a dataset

    A mesh can be constructed from this triangulation,
    for example using `Holoviews TriMesh <holoviews_>`_
    or `trimesh.Trimesh <trimesh_>`_.

    Parameters
    ----------
    dataset
        The dataset to triangulate

    Returns
    -------
    tuple of vertices, triangles, and cell indices.
        A tuple of three lists is returned,
        containing vertices, triangles, and cell indices respectively.

        Each vertex is a tuple of (x, y) or (lon, lat) coordinates.

        Each triangle is a tuple of three integers,
        indicating which vertices make up the triangle.

        The cell indices tie the triangles to the original cell polygon,
        allowing you to plot data on the triangle mesh.

    Examples
    --------

    Using holoviews_. Try this in an IPython notebook for nice visualisations:

    .. code-block:: python

        import emsarray
        import holoviews as hv
        from emsarray.operations import triangulate_dataset
        from holoviews import opts
        hv.extension('bokeh')

        # Triangulate the dataset
        dataset = emsarray.tuorial.open_dataset("austen")
        vertices, triangles, cell_indices = triangulate_dataset(dataset)

        # This takes a while to render
        mesh = hv.TriMesh((triangles, vertices))
        mesh

    Using trimesh_. This should pop up a new window to display the output:

    .. code-block:: python

        import emsarray
        import numpy as np
        import trimesh
        from emsarray.operations import triangulate_dataset

        dataset = emsarray.tutorial.open_dataset("gbr4")
        vertices, triangles, cell_indices = triangulate_dataset(dataset)
        # Trimesh expects 3D vertices.
        vertices = np.c_[vertices, np.zeros(len(vertices))]
        mesh = trimesh.Trimesh(vertices=vertices, faces=triangles)
        mesh.invert()  # Trimesh grids expect the opposite winding order

        depth = 1 - (dataset.data_vars["Mesh2_depth"].values / -200)
        depth_colour = np.c_[depth, depth, depth, np.ones_like(depth)] * 255
        mesh.visual.face_colors = depth_colour[cell_indices]
        mesh.show()

    .. _holoviews: https://holoviews.org/reference/elements/bokeh/TriMesh.html
    .. _trimesh: https://trimsh.org
    """
    polygons = dataset.ems.polygons

    # Getting all the vertices is easy - extract them from the polygons.
    # By going through a set, this will deduplicate the vertices.
    # Back to a list and we have a stable order
    vertices: List[Vertex] = list({
        vertex
        for polygon in polygons
        if polygon is not None
        for vertex in polygon.exterior.coords
    })

    # This maps between a vertex tuple and its index.
    # Vertex positions are (probably) floats. For grid datasets, where cells
    # are implicitly defined by their centres, be careful to compute cell
    # vertices in consistent ways. Float equality is tricky!
    vertex_indices = {vertex: index for index, vertex in enumerate(vertices)}

    # Each cell polygon needs to be triangulated,
    # while also recording the convention native index of the cell,
    # so that we can later correlate cell data with the triangles.
    polygons_with_index = [
        (polygon, index)
        for index, polygon in enumerate(polygons)
        if polygon is not None]
    triangles_with_index = list(
        (tuple(vertex_indices[vertex] for vertex in triangle_coords), dataset_index)
        for polygon, dataset_index in polygons_with_index
        for triangle_coords in _triangulate_polygon(polygon)
    )
    triangles: List[Triangle] = [tri for tri, index in triangles_with_index]  # type: ignore
    indices = [index for tri, index in triangles_with_index]

    return (vertices, triangles, indices)


def _triangulate_polygon(polygon: Polygon) -> List[Tuple[Vertex, Vertex, Vertex]]:
    """
    Triangulate a polygon.

    .. note::

        This currently only supports simple polygons - polygons that do not
        intersect themselves and do not have holes.

    Examples
    --------

    .. code-block:: python

        >>> polygon = Polygon([(0, 0), (2, 0), (2, 2), (1, 3), (0, 2), (0, 0)])
        >>> for triangle in triangulate_polygon(polygon):
        ...     print(triangle.wkt)
        POLYGON ((0 0, 2 0, 2 2, 0 0))
        POLYGON ((0 0, 2 2, 1 3, 0 0))
        POLYGON ((0 0, 1 3, 0 2, 0 0))

    See also
    --------
    :func:`triangulate_dataset`,
    `Polygon triangulation <https://en.wikipedia.org/wiki/Polygon_triangulation>`_
    """
    if not polygon.is_simple:
        raise ValueError("_triangulate_polygon only supports simple polygons")

    # This is the 'ear clipping' method of polygon triangulation.
    # In any simple polygon, there is guaranteed to be at least two 'ears'
    # - three neighbouring vertices whos diagonal is inside the polygon.
    # An ear can be clipped off the polygon, giving one triangle and a new,
    # smaller polygon. Repeat till the polygon is a triangle.
    #
    # This algorithm is not the best if given particularly bad polygons,
    # but for small convex polygons it is approximately linear in time.
    # Most polygons will be either squares, convex quadrilaterals, or convex
    # polygons.

    # Maintain a consistent winding order
    polygon = polygon.normalize()

    triangles: List[Tuple[Vertex, Vertex, Vertex]] = []
    # Note that shapely polygons with n vertices will be closed, and thus have
    # n+1 coordinates. We trim that superfluous coordinate off in the next line
    while len(polygon.exterior.coords) > 4:
        exterior = polygon.exterior
        coords = exterior.coords[:-1]
        # We try and make an ear from vertices (i, i+1, i+2)
        for i in range(len(coords) - 2):
            # If the diagonal between i and i+2 is within the larger polygon,
            # then this entire triangle is within the larger polygon
            vertices = [coords[i], coords[i + 2]]
            multipoint = MultiPoint(vertices)
            diagonal = LineString(vertices)
            if (
                diagonal.covered_by(polygon)
                # The intersection of the diagonal with the boundary should be two
                # points - the vertices in question. If three or more points are in
                # a line, the intersection will be something different. Removing
                # that ear will result in two disconnected polygons.
                and exterior.intersection(diagonal).equals(multipoint)
            ):
                triangles.append((coords[i], coords[i + 1], coords[i + 2]))
                polygon = Polygon(coords[:i + 1] + coords[i + 2:])
                break
        else:
            # According to the twos ear theorem, we should never reach this
            raise ValueError(
                f"Could not find interior diagonal for polygon! {polygon.wkt}")

    # The trimmed polygon is now a triangle. Add it to the list and we are done!

    triangles.append(cast(
        Tuple[Vertex, Vertex, Vertex],
        tuple(map(tuple, polygon.exterior.coords[:-1]))))
    return triangles
